unit Compile;

{
  Inno Setup
  Copyright (C) 1998-2001 Jordan Russell
  Portions by Martijn Laan
  For conditions of distribution and use, see LICENSE.TXT.

  Compiler

  $Id: Compile.pas,v 1.60 2001/12/03 19:10:40 jr Exp $
}

{$I VERSION.INC}

interface

uses
  Windows, SysUtils, CompInt;

function ISCompileScript (const Params: TCompileScriptParams;
  const PropagateExceptions: Boolean): Integer;
function ISGetVersion: PCompilerVersionInfo;

type
  EISCompileError = class(Exception)
  public
    Filename: String;
    LineNumber: Integer;
  end;

implementation

uses
  Commctrl, Consts,
  Classes, IniFiles, TypInfo,
  MBCSUtil, CmnFunc2, Struct, CompMsgs, SetupEnt, zlib, bzlib, MsgIDs;

type
  TWizardStyle = (wsClassic, wsModern);

  TBreakStringRec = record
    ParamName: String;
    ParamData: String;
  end;
  PBreakStringArray = ^TBreakStringArray;
  TBreakStringArray = array[0..15] of TBreakStringRec;

  TParamInfo = record
    Name: PChar;
    Flags: set of (piNoEmpty, piNoQuotes);
  end;

  TEnumIniSectionProc = procedure(const Line: PChar; const Ext: Integer) of object;

  TSetupSectionDirectives = (
    ssAdminPrivilegesRequired,
    ssAllowNoIcons,
    ssAllowRootDirectory,
    ssAllowUNCPath,
    ssAlwaysCreateUninstallIcon,
    ssAlwaysRestart,
    ssAlwaysShowComponentsList,
    ssAlwaysShowDirOnReadyPage,
    ssAlwaysShowGroupOnReadyPage,
    ssAlwaysUsePersonalGroup,
    ssAppCopyright,
    ssAppId,
    ssAppMutex,
    ssAppName,
    ssAppPublisher,
    ssAppPublisherURL,
    ssAppSupportURL,
    ssAppUpdatesURL,
    ssAppVerName,
    ssAppVersion,
    ssBackColor,
    ssBackColor2,
    ssBackColorDirection,
    ssBackSolid,
    ssBits,
    ssChangesAssociations,
    ssCompression,
    ssCompressLevel,
    ssCreateAppDir,
    ssCreateUninstallRegKey,
    ssDefaultDirName,
    ssDefaultGroupName,
    ssDirExistsWarning,
    ssDisableAppendDir,
    ssDisableDirExistsWarning,
    ssDisableDirPage,
    ssDisableFinishedPage,
    ssDisableProgramGroupPage,
    ssDisableReadyMemo,
    ssDisableReadyPage,
    ssDisableStartupPrompt,
    ssDiskClusterSize,
    ssDiskSize,
    ssDiskSpanning,
    ssDontMergeDuplicateFiles,
    ssEnableDirDoesntExistWarning,
    ssExtraDiskSpaceRequired,
    ssInfoAfterFile,
    ssInfoBeforeFile,
    ssInternalCompressLevel,
    ssInstallMode,
    ssFlatComponentsList,
    ssLicenseFile,
    ssMessagesFile,
    ssMinVersion,
    ssOnlyBelowVersion,
    ssOutputBaseFilename,
    ssOutputDir,
    ssOverwriteUninstRegEntries,
    ssPassword,
    ssReserveBytes,
    ssShowComponentSizes,
    ssSourceDir,
    ssUpdateUninstallLogAppName,
    ssUninstallable,
    ssUninstallDisplayIcon,
    ssUninstallDisplayName,
    ssUninstallFilesDir,
    ssUninstallIconName,
    ssUninstallLogMode,
    ssUninstallStyle,
    ssUsePreviousAppDir,
    ssUsePreviousGroup,
    ssUsePreviousSetupType,
    ssUsePreviousTasks,
    ssUseSetupLdr,
    ssWindowResizable,
    ssWindowShowCaption,
    ssWindowStartMaximized,
    ssWindowVisible,
    ssWizardImageBackColor,
    ssWizardImageFile,
    ssWizardSmallImageBackColor,
    ssWizardSmallImageFile,
    ssWizardStyle);
  TLangOptionsSectionDirectives = (
    lsCopyrightFontName,
    lsCopyrightFontSize,
    lsDialogFontName,
    lsDialogFontSize,
    lsDialogFontStandardHeight,
    lsLanguageID,
    lsLanguageName,
    lsTitleFontName,
    lsTitleFontSize,
    lsWelcomeFontName,
    lsWelcomeFontSize);

  TAllowedConst = (acOldData, acBreak);
  TAllowedConsts = set of TAllowedConst;

  TSetupCompiler = class
  private
    TypeEntries,
    ComponentEntries,
    TaskEntries,
    DirEntries,
    FileEntries,
    FileLocationEntries,
    IconEntries,
    IniEntries,
    RegistryEntries,
    InstallDeleteEntries,
    UninstallDeleteEntries,
    RunEntries,
    UninstallRunEntries: TList;

    FileLocationEntryFilenames, WarningsList: TStringList;

    Messages: array[TSetupMessageID] of PChar;

    WizardStyle: TWizardStyle;

    OutputDir, OutputBaseFilename, ExeFilename: String;
    CompressWithBzip: Boolean;
    InternalCompressLevel, CompressLevel: Integer;
    DontMergeDuplicateFiles: Boolean;

    SetupHeader: ^TSetupHeader;
    LangOptions: TSetupLangOptions;

    SetupDirectiveLines: array[TSetupSectionDirectives] of Integer;
    LangOptionsDirectivesSpecified: set of TLangOptionsSectionDirectives;
    UseSetupLdr, DiskSpanning, HasRegSvr, HasRestart, BackSolid: Boolean;
    DiskTotalBytes, DiskClusterSize, ReserveBytes: Longint;
    MessagesFile, LicenseFile, InfoBeforeFile, InfoAfterFile, WizardImageFile: String;
    WizardSmallImageFile: String;

    LineNumber, AbortLineNumber: Integer;
    AbortFilename: String;

    procedure AddStatus (const S: String);
    procedure AbortCompile (const Msg: String);
    procedure AbortCompileFmt (const Msg: String; const Args: array of const);
    procedure AbortCompileOnLine (const Msg: String);
    procedure AbortCompileOnLineFmt (const Msg: String;
      const Args: array of const);
    procedure AbortCompileParamError (const Msg, ParamName: String);
    function PrependSourceDirName (const Filename: String): String;
    procedure BreakString (S: PChar; var Output: TBreakStringArray);
    procedure CallIdleProc;
    procedure DoCallback (const Code: Integer; var Data: TCompilerCallbackData);
    procedure EnumIniSection (const EnumProc: TEnumIniSectionProc;
      const SectionName: String; const Ext: Integer; const Verbose: Boolean;
      const ReadFromFile: TextFile);
    function MinVersionLessThan4 (const MinVersion: TSetupVersionData): Boolean;
    procedure CheckDirConst (const S: String; const MinVersion: TSetupVersionData;
      const AllowedConsts: TAllowedConsts);
    function CompareParamName (const S: TBreakStringRec;
      const ParamInfo: array of TParamInfo;
      var ParamNamesFound: array of Boolean): Integer;
    procedure EnumTypes (const Line: PChar; const Ext: Integer);
    procedure EnumComponents (const Line: PChar; const Ext: Integer);
    procedure EnumTasks (const Line: PChar; const Ext: Integer);
    procedure EnumDirs (const Line: PChar; const Ext: Integer);
    procedure EnumIcons (const Line: PChar; const Ext: Integer);
    procedure EnumINI (const Line: PChar; const Ext: Integer);
    procedure EnumLangOptions (const Line: PChar; const Ext: Integer);
    procedure EnumRegistry (const Line: PChar; const Ext: Integer);
    procedure EnumDelete (const Line: PChar; const Ext: Integer);
    procedure EnumFiles (const Line: PChar; const Ext: Integer);
    procedure EnumRun (const Line: PChar; const Ext: Integer);
    procedure EnumSetup (const Line: PChar; const Ext: Integer);
    procedure EnumMessages (const Line: PChar; const Ext: Integer);
    procedure InitBzipDLL;
    procedure InitLangOptions (var ALangOptions: TSetupLangOptions);
    procedure ReadMessages;
    procedure SeparateDirective (const Line: PChar; var Key, Value: String);
  public
    AppData: Longint;
    CallbackProc: TCompilerCallbackProc;
    CompilerDir, SourceDir: String;
    constructor Create (AOwner: TComponent);
    destructor Destroy; override;
    procedure Compile;
  end;

var
  BzipInitialized: Boolean;

const
  ParamCommonFlags = 'Flags';
  ParamCommonComponents = 'Components';
  ParamCommonTasks = 'Tasks';
  ParamCommonMinVersion = 'MinVersion';
  ParamCommonOnlyBelowVersion = 'OnlyBelowVersion';

  DefaultOptionsModern = [shCreateAppDir, shUninstallable,
    shWindowStartMaximized, shWindowShowCaption, shWindowResizable,
    shCreateUninstallRegKey, shUsePreviousAppDir, shUsePreviousGroup, shUsePreviousSetupType,
    shAlwaysShowComponentsList, shFlatComponentsList, shShowComponentSizes,
    shUsePreviousTasks, shUpdateUninstallLogAppName, shAllowUNCPath];

  DefaultWizardImageBackColorClassic = $808000;
  DefaultWizardImageBackColorModern = $400000;

  DefaultUninstallStyleClassic = usClassic;
  DefaultUninstallStyleModern = usModern;

  DefaultWizardImageFileClassic = 'compiler:WIZCLASSICIMAGE.BMP';
  DefaultWizardImageFileModern = 'compiler:WIZMODERNIMAGE.BMP';

  DefaultTypeEntryNames: array[0..2] of PChar = ('full', 'compact', 'custom');

  MaxDiskSize = 2100000000;

type
  TColor = $7FFFFFFF-1..$7FFFFFFF;

const
  clScrollBar = TColor(COLOR_SCROLLBAR or $80000000);
  clBackground = TColor(COLOR_BACKGROUND or $80000000);
  clActiveCaption = TColor(COLOR_ACTIVECAPTION or $80000000);
  clInactiveCaption = TColor(COLOR_INACTIVECAPTION or $80000000);
  clMenu = TColor(COLOR_MENU or $80000000);
  clWindow = TColor(COLOR_WINDOW or $80000000);
  clWindowFrame = TColor(COLOR_WINDOWFRAME or $80000000);
  clMenuText = TColor(COLOR_MENUTEXT or $80000000);
  clWindowText = TColor(COLOR_WINDOWTEXT or $80000000);
  clCaptionText = TColor(COLOR_CAPTIONTEXT or $80000000);
  clActiveBorder = TColor(COLOR_ACTIVEBORDER or $80000000);
  clInactiveBorder = TColor(COLOR_INACTIVEBORDER or $80000000);
  clAppWorkSpace = TColor(COLOR_APPWORKSPACE or $80000000);
  clHighlight = TColor(COLOR_HIGHLIGHT or $80000000);
  clHighlightText = TColor(COLOR_HIGHLIGHTTEXT or $80000000);
  clBtnFace = TColor(COLOR_BTNFACE or $80000000);
  clBtnShadow = TColor(COLOR_BTNSHADOW or $80000000);
  clGrayText = TColor(COLOR_GRAYTEXT or $80000000);
  clBtnText = TColor(COLOR_BTNTEXT or $80000000);
  clInactiveCaptionText = TColor(COLOR_INACTIVECAPTIONTEXT or $80000000);
  clBtnHighlight = TColor(COLOR_BTNHIGHLIGHT or $80000000);
  cl3DDkShadow = TColor(COLOR_3DDKSHADOW or $80000000);
  cl3DLight = TColor(COLOR_3DLIGHT or $80000000);
  clInfoText = TColor(COLOR_INFOTEXT or $80000000);
  clInfoBk = TColor(COLOR_INFOBK or $80000000);

  clBlack = TColor($000000);
  clMaroon = TColor($000080);
  clGreen = TColor($008000);
  clOlive = TColor($008080);
  clNavy = TColor($800000);
  clPurple = TColor($800080);
  clTeal = TColor($808000);
  clGray = TColor($808080);
  clSilver = TColor($C0C0C0);
  clRed = TColor($0000FF);
  clLime = TColor($00FF00);
  clYellow = TColor($00FFFF);
  clBlue = TColor($FF0000);
  clFuchsia = TColor($FF00FF);
  clAqua = TColor($FFFF00);
  clLtGray = TColor($C0C0C0);
  clDkGray = TColor($808080);
  clWhite = TColor($FFFFFF);
  clNone = TColor($1FFFFFFF);
  clDefault = TColor($20000000);

type
  TColorEntry = record
    Value: TColor;
    Name: string;
  end;

const
  Colors: array[0..41] of TColorEntry = (
    (Value: clBlack; Name: 'clBlack'),
    (Value: clMaroon; Name: 'clMaroon'),
    (Value: clGreen; Name: 'clGreen'),
    (Value: clOlive; Name: 'clOlive'),
    (Value: clNavy; Name: 'clNavy'),
    (Value: clPurple; Name: 'clPurple'),
    (Value: clTeal; Name: 'clTeal'),
    (Value: clGray; Name: 'clGray'),
    (Value: clSilver; Name: 'clSilver'),
    (Value: clRed; Name: 'clRed'),
    (Value: clLime; Name: 'clLime'),
    (Value: clYellow; Name: 'clYellow'),
    (Value: clBlue; Name: 'clBlue'),
    (Value: clFuchsia; Name: 'clFuchsia'),
    (Value: clAqua; Name: 'clAqua'),
    (Value: clWhite; Name: 'clWhite'),
    (Value: clScrollBar; Name: 'clScrollBar'),
    (Value: clBackground; Name: 'clBackground'),
    (Value: clActiveCaption; Name: 'clActiveCaption'),
    (Value: clInactiveCaption; Name: 'clInactiveCaption'),
    (Value: clMenu; Name: 'clMenu'),
    (Value: clWindow; Name: 'clWindow'),
    (Value: clWindowFrame; Name: 'clWindowFrame'),
    (Value: clMenuText; Name: 'clMenuText'),
    (Value: clWindowText; Name: 'clWindowText'),
    (Value: clCaptionText; Name: 'clCaptionText'),
    (Value: clActiveBorder; Name: 'clActiveBorder'),
    (Value: clInactiveBorder; Name: 'clInactiveBorder'),
    (Value: clAppWorkSpace; Name: 'clAppWorkSpace'),
    (Value: clHighlight; Name: 'clHighlight'),
    (Value: clHighlightText; Name: 'clHighlightText'),
    (Value: clBtnFace; Name: 'clBtnFace'),
    (Value: clBtnShadow; Name: 'clBtnShadow'),
    (Value: clGrayText; Name: 'clGrayText'),
    (Value: clBtnText; Name: 'clBtnText'),
    (Value: clInactiveCaptionText; Name: 'clInactiveCaptionText'),
    (Value: clBtnHighlight; Name: 'clBtnHighlight'),
    (Value: cl3DDkShadow; Name: 'cl3DDkShadow'),
    (Value: cl3DLight; Name: 'cl3DLight'),
    (Value: clInfoText; Name: 'clInfoText'),
    (Value: clInfoBk; Name: 'clInfoBk'),
    (Value: clNone; Name: 'clNone'));

function IdentToColor(const Ident: string; var Color: Longint): Boolean;
var
  I: Integer;
begin
  for I := Low(Colors) to High(Colors) do
    if CompareText(Colors[I].Name, Ident) = 0 then
    begin
      Result := True;
      Color := Longint(Colors[I].Value);
      Exit;
    end;
  Result := False;
end;

function StringToColor(const S: string): TColor;
begin
  if not IdentToColor(S, Longint(Result)) then
    Result := TColor(StrToInt(S));
end;

function AddPeriod (const S: String): String;
begin
  Result := S;
  if (Result <> '') and (AnsiLastChar(Result)^ > '.') then
    Result := Result + '.';
end;

function IsRelativePath (const Filename: String): Boolean;
var
  L: Integer;
begin
  Result := True;
  L := Length(Filename);
  if ((L >= 1) and (Filename[1] = '\')) or
     ((L >= 2) and (Filename[1] in ['A'..'Z', 'a'..'z']) and (Filename[2] = ':')) then
    Result := False;
end;

function GetSelfFilename: String;
{ Returns Filename of the calling DLL or application. (ParamStr(0) can only
  return the filename of the calling application.) }
var
  Buf: array[0..MAX_PATH-1] of Char;
begin
  SetString (Result, Buf, GetModuleFileName(HInstance, Buf, SizeOf(Buf)))
end;

function FileSizeAndCRCIs (const Filename: String; const Size: Longint;
  const CRC: Longint): Boolean;
var
  F: File;
  Buf: String;
begin
  Result := False;
  {$IFOPT I+}
    {$DEFINE HAD_OPT_I}
    {$I-}
  {$ENDIF}
  AssignFile (F, Filename);
  Reset (F, 1);
  if IOResult <> 0 then Exit;
  try
    if FileSize(F) <> Size then
      Exit;
    SetLength (Buf, Size);
    BlockRead (F, Buf[1], Size);
    if IOResult <> 0 then Exit;
    if GetCRC32(Buf[1], Size) = CRC then
      Result := True;
  finally
    CloseFile (F);
  end;
  {$IFDEF HAD_OPT_I}
    {$UNDEF HAD_OPT_I}
    {$I+}
  {$ENDIF}
end;


{ TSetupCompiler }

constructor TSetupCompiler.Create (AOwner: TComponent);
begin
  inherited Create;
  TypeEntries := TList.Create;
  ComponentEntries := TList.Create;
  TaskEntries := TList.Create;
  DirEntries := TList.Create;
  FileEntries := TList.Create;
  FileLocationEntries := TList.Create;
  IconEntries := TList.Create;
  IniEntries := TList.Create;
  RegistryEntries := TList.Create;
  InstallDeleteEntries := TList.Create;
  UninstallDeleteEntries := TList.Create;
  RunEntries := TList.Create;
  UninstallRunEntries := TList.Create;
  FileLocationEntryFilenames := TStringList.Create;
  WarningsList := TStringList.Create;
end;

destructor TSetupCompiler.Destroy;
begin
  WarningsList.Free;
  FileLocationEntryFilenames.Free;
  UninstallRunEntries.Free;
  RunEntries.Free;
  UninstallDeleteEntries.Free;
  InstallDeleteEntries.Free;
  RegistryEntries.Free;
  IniEntries.Free;
  IconEntries.Free;
  FileLocationEntries.Free;
  FileEntries.Free;
  DirEntries.Free;
  TaskEntries.Free;
  ComponentEntries.Free;
  TypeEntries.Free;
  inherited Destroy;
end;

procedure TSetupCompiler.InitBzipDLL;
var
  M: HMODULE;
begin
  if BzipInitialized then
    Exit;
  M := LoadLibrary(PChar(CompilerDir + 'isbzip.dll'));
  if M = 0 then
    AbortCompileFmt ('Failed to load isbzip.dll (%d)', [GetLastError]);
  if not BZInitCompressFunctions(M) then
    AbortCompile ('Failed to get address of functions in isbzip.dll');
  BzipInitialized := True;
end;

procedure TSetupCompiler.DoCallback (const Code: Integer;
  var Data: TCompilerCallbackData);
begin
  if CallbackProc(Code, Data, AppData) = iscrRequestAbort then
    Abort;
end;

procedure TSetupCompiler.CallIdleProc;
begin
  DoCallback (iscbNotifyIdle, TCompilerCallbackData(nil^));
end;

procedure TSetupCompiler.EnumIniSection (const EnumProc: TEnumIniSectionProc;
  const SectionName: String; const Ext: Integer; const Verbose: Boolean;
  const ReadFromFile: TextFile);
var
  Data: TCompilerCallbackData;
  Reset: Boolean;
  FoundSection: Boolean;
  B, L, LastSection: String;
  SaveLineNumber, I: Integer;
begin
  SaveLineNumber := LineNumber;
  try
    FoundSection := False;
    LineNumber := 0;
    Reset := True;
    LastSection := '';
    while True do begin
      Inc (LineNumber);
      if @ReadFromFile = nil then begin
        Data.Reset := Reset;
        Data.LineRead := nil;
        DoCallback (iscbReadScript, Data);
        if Data.LineRead = nil then
          Break;
        B := Data.LineRead;
      end
      else begin
        if Reset then
          System.Reset (ReadFromFile);
        if Eof(ReadFromFile) then
          Break;
        Readln (ReadFromFile, B);
      end;
      Reset := False;
      L := Trim(B);
      { Check for blank lines or comments }
      if (L = '') or (L[1] = ';') then Continue;
      if L[1] = '[' then begin
        I := Pos(']', L);
        if I < 3 then
          AbortCompileOnLine (SCompilerSectionTagInvalid);
        L := Copy(L, 2, I-2);
        if L[1] = '/' then begin
          L := Copy(L, 2, Maxint);
          if (LastSection = '') or (CompareText(L, LastSection) <> 0) then
            AbortCompileOnLineFmt (SCompilerSectionBadEndTag, [L]);
          LastSection := '';
          if FoundSection then
            { end tag was for the section we were enumerating }
            Break;
        end
        else begin
          if FoundSection then
            { no end tag, but a new section has been encountered }
            Break;
          if CompareText(L, SectionName) = 0 then
            FoundSection := True;
          LastSection := L;
        end;
      end
      else begin
        if not FoundSection then Continue;  { not on the right section }
        if Verbose then
          AddStatus (Format(SCompilerStatusReadSectionLine, [SectionName, LineNumber]));
        EnumProc (PChar(B), Ext);
      end;
    end;
  finally
    LineNumber := SaveLineNumber;
  end;
end;

procedure TSetupCompiler.BreakString (S: PChar; var Output: TBreakStringArray);
var
  ColonPos, SemicolonPos, QuotePos, P, P2: PChar;
  ParamName, Data: String;
  QuoteFound, FirstQuoteFound, LastQuoteFound, AddChar, FirstNonspaceFound: Boolean;
  CurParm, Len, I: Integer;
begin
  CurParm := 0;
  while (S <> nil) and (CurParm <= High(TBreakStringArray)) do begin
    ColonPos := StrScan(S, ':');
    if ColonPos = nil then
      ParamName := StrPas(S)
    else
      SetString (ParamName, S, ColonPos-S);
    ParamName := Trim(ParamName);
    if ParamName = '' then Break;
    if ColonPos = nil then
      AbortCompileOnLineFmt (SCompilerParamHasNoValue, [ParamName]);
    S := ColonPos + 1;
    SemicolonPos := StrScan(S, ';');
    QuotePos := StrScan(S, '"');
    QuoteFound := QuotePos <> nil;
    if QuoteFound and (SemicolonPos <> nil) and (QuotePos > SemicolonPos) then
      QuoteFound := False;
    if not QuoteFound then begin
      Data := '';
      P := S;
      if SemicolonPos <> nil then
        P2 := SemicolonPos
      else
        P2 := StrEnd(S);
      FirstNonspaceFound := False;
      Len := 0;
      I := 0;
      while P < P2 do begin
        if (P^ <> ' ') or FirstNonspaceFound then begin
          FirstNonspaceFound := True;
          Data := Data + P^;
          Inc (I);
          if P^ <> ' ' then Len := I;
        end;
        Inc (P);
      end;
      SetLength (Data, Len);
    end
    else begin
      Data := '';
      SemicolonPos := nil;
      P := S;
      FirstQuoteFound := False;
      LastQuoteFound := False;
      while P^ <> #0 do begin
        AddChar := False;
        case P^ of
          ' ': AddChar := FirstQuoteFound;
          '"': if not FirstQuoteFound then
                 FirstQuoteFound := True
               else begin
                 Inc (P);
                 if P^ = '"' then
                   AddChar := True
                 else begin
                   LastQuoteFound := True;
                   while P^ <> #0 do begin
                     case P^ of
                       ' ': ;
                       ';': begin
                              SemicolonPos := P;
                              Break;
                            end;
                     else
                       AbortCompileOnLineFmt (SCompilerParamQuoteError, [ParamName]);
                     end;
                     Inc (P);
                   end;
                   Break;
                 end;
               end;
        else
          if not FirstQuoteFound then
            AbortCompileOnLineFmt (SCompilerParamQuoteError, [ParamName]);
          AddChar := True;
        end;
        if AddChar then
          Data := Data + P^;
        Inc (P);
      end;
      if not LastQuoteFound then
        AbortCompileOnLineFmt (SCompilerParamMissingClosingQuote, [ParamName]);
    end;
    S := SemicolonPos;
    if S <> nil then Inc (S);
    Output[CurParm].ParamName := ParamName;
    Output[CurParm].ParamData := Data;
    Inc (CurParm);
  end;
end;

procedure TSetupCompiler.AddStatus (const S: String);
var
  Data: TCompilerCallbackData;
begin
  Data.StatusMsg := PChar(S);
  CallbackProc (iscbNotifyStatus, Data, AppData);
end;

procedure TSetupCompiler.AbortCompile (const Msg: String);
var
  E: EISCompileError;
begin
  E := EISCompileError.Create(Msg);
  E.Filename := AbortFilename;
  E.LineNumber := AbortLineNumber;
  raise E;
end;

procedure TSetupCompiler.AbortCompileFmt (const Msg: String; const Args: array of const);
begin
  AbortCompile (Format(Msg, Args));
end;

procedure TSetupCompiler.AbortCompileOnLine (const Msg: String);
begin
  AbortLineNumber := LineNumber;
  AbortCompile (Msg);
end;

procedure TSetupCompiler.AbortCompileOnLineFmt (const Msg: String;
  const Args: array of const);
begin
  AbortCompileOnLine (Format(Msg, Args));
end;

procedure TSetupCompiler.AbortCompileParamError (const Msg, ParamName: String);
begin
  AbortCompileOnLineFmt (Msg, [ParamName]);
end;

function TSetupCompiler.PrependSourceDirName (const Filename: String): String;
begin
  if CompareText(Copy(Filename, 1, 9), 'compiler:') = 0 then
    Result := CompilerDir + Copy(Filename, 10, Maxint)
  else begin
    if (Filename = '') or not IsRelativePath(Filename) then
      Result := Filename
    else
      Result := SourceDir + Filename;
  end;
end;

function TSetupCompiler.MinVersionLessThan4 (const MinVersion: TSetupVersionData): Boolean;
begin
  with MinVersion do
    Result := ((WinVersion <> 0) and (WinVersion < $04000000)) or
      ((NTVersion <> 0) and (NTVersion < $04000000));
end;

function MinVersionErrorMessage (const Param: Boolean;
  const WinVersion, NTVersion: Word): String;

  function VerToStr (Ver: Cardinal; ServicePack: Word): String;
  var
    Digits: Integer;
  begin
    with TSetupVersionDataVersion(Ver) do begin
      Digits := 2;
      if Minor mod 10 = 0 then begin
        Dec (Digits);
        Minor := Minor div 10;
      end;
      FmtStr (Result, '%d.%.*d', [Major, Digits, Minor]);
      if Build <> 0 then
        Result := Result + Format('.%d', [Build]);
      if ServicePack <> 0 then begin
        Result := Result + ' Service Pack ' + IntToStr(Hi(ServicePack));
        if Lo(ServicePack) <> 0 then
          Result := Result + Format('.%d', [Lo(ServicePack)]);
      end;
    end;
  end;

var
  WinVer, NTVer: String;
begin
  WinVer := VerToStr(WinVersion shl 16, 0);
  NTVer := VerToStr(NTVersion shl 16, 0);
  if not Param then
    FmtStr (Result, SCompilerMinVersionError, [WinVer, NTVer, WinVer, NTVer])
  else
    FmtStr (Result, SCompilerMinVersionErrorParam, [WinVer, NTVer,
      WinVer, NTVer, WinVer, NTVer]);
end;

procedure TSetupCompiler.CheckDirConst (const S: String;
  const MinVersion: TSetupVersionData; const AllowedConsts: TAllowedConsts);

  function CheckRegConst (C: String): Boolean;
  { based on ExpandRegConst in Main.pas }
  type
    TKeyNameConst = packed record
      KeyName: String;
      KeyConst: HKEY;
    end;
  const
    KeyNameConsts: array[0..4] of TKeyNameConst = (
      (KeyName: 'HKCR'; KeyConst: HKEY_CLASSES_ROOT),
      (KeyName: 'HKCU'; KeyConst: HKEY_CURRENT_USER),
      (KeyName: 'HKLM'; KeyConst: HKEY_LOCAL_MACHINE),
      (KeyName: 'HKU';  KeyConst: HKEY_USERS),
      (KeyName: 'HKCC'; KeyConst: HKEY_CURRENT_CONFIG));
  var
    Z, Subkey, Value, Default: String;
    I, J: Integer;
    RootKey: HKEY;
  begin
    Delete (C, 1, 4);  { skip past 'reg:' }
    I := ConstPos('\', C);
    if I <> 0 then begin
      Z := Copy(C, 1, I-1);
      if Z <> '' then begin
        RootKey := 0;
        for J := Low(KeyNameConsts) to High(KeyNameConsts) do
          if CompareText(KeyNameConsts[J].KeyName, Z) = 0 then begin
            RootKey := KeyNameConsts[J].KeyConst;
            Break;
          end;
        if RootKey <> 0 then begin
          Z := Copy(C, I+1, Maxint);
          I := ConstPos('|', Z);  { check for a 'default' data }
          if I = 0 then
            I := Length(Z)+1;
          Default := Copy(Z, I+1, Maxint);
          SetLength (Z, I-1);
          I := ConstPos(',', Z);  { comma separates subkey and value }
          if I <> 0 then begin
            Subkey := Copy(Z, 1, I-1);
            Value := Copy(Z, I+1, Maxint);
            if ConvertConstPercentStr(Subkey) and ConvertConstPercentStr(Value) and
               ConvertConstPercentStr(Default) then begin
              CheckDirConst (Subkey, MinVersion, AllowedConsts);
              CheckDirConst (Value, MinVersion, AllowedConsts);
              CheckDirConst (Default, MinVersion, AllowedConsts);
              Result := True;
              Exit;
            end;
          end;
        end;
      end;
    end;
    { it will only reach here if there was a parsing error }
    Result := False;
  end;

  function CheckIniConst (C: String): Boolean;
  { based on ExpandIniConst in Main.pas }
  var
    Z, Filename, Section, Key, Default: String;
    I: Integer;
  begin
    Delete (C, 1, 4);  { skip past 'ini:' }
    I := ConstPos(',', C);
    if I <> 0 then begin
      Z := Copy(C, 1, I-1);
      if Z <> '' then begin
        Filename := Z;
        Z := Copy(C, I+1, Maxint);
        I := ConstPos('|', Z);  { check for a 'default' data }
        if I = 0 then
          I := Length(Z)+1;
        Default := Copy(Z, I+1, Maxint);
        SetLength (Z, I-1);
        I := ConstPos(',', Z);  { comma separates section and key }
        if I <> 0 then begin
          Section := Copy(Z, 1, I-1);
          Key := Copy(Z, I+1, Maxint);
          if ConvertConstPercentStr(Filename) and ConvertConstPercentStr(Section) and
             ConvertConstPercentStr(Key) and ConvertConstPercentStr(Default) then begin
            CheckDirConst (Filename, MinVersion, AllowedConsts);
            CheckDirConst (Section, MinVersion, AllowedConsts);
            CheckDirConst (Key, MinVersion, AllowedConsts);
            CheckDirConst (Default, MinVersion, AllowedConsts);
            Result := True;
            Exit;
          end;
        end;
      end;
    end;
    { it will only reach here if there was a parsing error }
    Result := False;
  end;

  function CheckParamConst (C: String): Boolean;
  var
    Z, Param, Default: String;
    I: Integer;
  begin
    Delete (C, 1, 6);  { skip past 'param:' }
    Z := C;
    I := ConstPos('|', Z);  { check for a 'default' data }
    if I = 0 then
      I := Length(Z)+1;
    Default := Copy(Z, I+1, Maxint);
    SetLength (Z, I-1);
    Param := Z;
    if ConvertConstPercentStr(Param) and ConvertConstPercentStr(Default) then begin
      CheckDirConst (Param, MinVersion, AllowedConsts);
      CheckDirConst (Default, MinVersion, AllowedConsts);
      Result := True;
      Exit;
    end;
    { it will only reach here if there was a parsing error }
    Result := False;
  end;

const
  Consts: array[0..15] of PChar = (
    'src', 'srcexe', 'tmp', 'app', 'win', 'sys', 'sd', 'groupname', 'fonts',
    'hwnd', 'pf', 'cf', 'computername', 'dao', 'username', 'wizardhwnd');
  ShellFolderConsts: array[0..16] of PChar = (
    'group', 'userdesktop', 'userstartmenu', 'userprograms', 'userstartup',
    'commondesktop', 'commonstartmenu', 'commonprograms', 'commonstartup',
    'sendto', 'userappdata', 'userdocs', 'commonappdata', 'commondocs',
    'usertemplates', 'commontemplates', 'localappdata');
  ShellFolderConsts98: array[0..1] of PChar = ('userfavorites', 'commonfavorites');
  AllowedConstsNames: array[TAllowedConst] of PChar = (
    'olddata', 'break');
var
  NoShellConsts, NotWin98orNT4: Boolean;
  I, Start, K: Integer;
  C: TAllowedConst;
  Cnst: String;
label 1;
begin
  NoShellConsts := MinVersionLessThan4(MinVersion);
  with MinVersion do
    NotWin98orNT4 := ((WinVersion <> 0) and (WinVersion < $040A0000)) or
      ((NTVersion <> 0) and (NTVersion < $04000000));
  I := 1;
  while I <= Length(S) do begin
    if S[I] = '{' then begin
      if (I < Length(S)) and (S[I+1] = '{') then
        Inc (I)
      else begin
        Start := I;
        { Find the closing brace, skipping over any embedded constants }
        I := SkipPastConst(S, I);
        if I = 0 then  { unclosed constant? }
          AbortCompileOnLineFmt (SCompilerUnterminatedConst, [Copy(S, Start+1, Maxint)]);
        Dec (I);  { 'I' now points to the closing brace }

        { Now check the constant }
        Cnst := Copy(S, Start+1, I-(Start+1));
        if Cnst <> '' then begin
          if (Cnst[1] = '%') or (Cnst = '\') then
            goto 1;
          if Copy(Cnst, 1, 4) = 'reg:' then begin
            if not CheckRegConst(Cnst) then
              AbortCompileOnLineFmt (SCompilerBadRegConst, [Cnst]);
            goto 1;
          end;
          if Copy(Cnst, 1, 4) = 'ini:' then begin
            if not CheckIniConst(Cnst) then
              AbortCompileOnLineFmt (SCompilerBadIniConst, [Cnst]);
            goto 1;
          end;
          if Copy(Cnst, 1, 6) = 'param:' then begin
            if not CheckParamConst(Cnst) then
              AbortCompileOnLineFmt (SCompilerBadParamConst, [Cnst]);
            goto 1;
          end;
          for K := Low(Consts) to High(Consts) do
            if Cnst = StrPas(Consts[K]) then
              goto 1;
          for K := Low(ShellFolderConsts) to High(ShellFolderConsts) do
            if Cnst = StrPas(ShellFolderConsts[K]) then begin
              if NoShellConsts then
                AbortCompileOnLineFmt (SCompilerConstUsed + SNewLine2 +
                  MinVersionErrorMessage(True, $400, $400), [Cnst]);
              goto 1;
            end;
          for K := Low(ShellFolderConsts98) to High(ShellFolderConsts98) do
            if Cnst = StrPas(ShellFolderConsts98[K]) then begin
              if NotWin98orNT4 then
                AbortCompileOnLineFmt (SCompilerConstUsed + SNewLine2 +
                  MinVersionErrorMessage(True, $40A, $400), [Cnst]);
              goto 1;
            end;
          for C := Low(C) to High(C) do
            if Cnst = StrPas(AllowedConstsNames[C]) then begin
              if not(C in AllowedConsts) then
                AbortCompileOnLineFmt (SCompilerConstCannotUse, [Cnst]);
              goto 1;
            end;
         end;
         AbortCompileOnLineFmt (SCompilerUnknownConst, [Cnst]);

      1:{ Constant is OK }
      end;
    end
    else if S[I] in LeadBytes then
      Inc (I);
    Inc (I);
  end;
end;

function TSetupCompiler.CompareParamName (const S: TBreakStringRec;
  const ParamInfo: array of TParamInfo;
  var ParamNamesFound: array of Boolean): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := 0 to High(ParamInfo) do
    if StrIComp(ParamInfo[I].Name, PChar(S.ParamName)) = 0 then begin
      Result := I;
      if ParamNamesFound[I] then
        AbortCompileParamError (SCompilerParamDuplicated, StrPas(ParamInfo[I].Name));
      ParamNamesFound[I] := True;
      if (piNoEmpty in ParamInfo[I].Flags) and (S.ParamData = '') then
        AbortCompileParamError (SCompilerParamEmpty2, StrPas(ParamInfo[I].Name));
      if (piNoQuotes in ParamInfo[I].Flags) and (Pos('"', S.ParamData) <> 0) then
        AbortCompileParamError (SCompilerParamNoQuotes2, StrPas(ParamInfo[I].Name));
      Exit;
    end;
  { Unknown parameter }
  AbortCompileOnLineFmt (SCompilerParamUnknownParam, [S.ParamName]);
end;

function ExtractStr (var S: String): String;
var
  I: Integer;
begin
  I := Pos(' ', S);
  if I = 0 then I := Length(S)+1;
  Result := Trim(Copy(S, 1, I-1));
  S := Trim(Copy(S, I+1, Maxint));
end;

function ExtractFlag (var S: String; const FlagStrs: array of PChar): Integer;
var
  I: Integer;
  F: String;
begin
  F := ExtractStr(S);
  if F = '' then begin
    Result := -2;
    Exit;
  end;

  Result := -1;
  for I := 0 to High(FlagStrs) do
    if StrIComp(FlagStrs[I], PChar(F)) = 0 then begin
      Result := I;
      Break;
    end;
end;

function ExtractType(var S: String; const TypeEntries: TList): Integer;
var
  I: Integer;
  F: String;
begin
  F := ExtractStr(S);
  if F = '' then begin
    Result := -2;
    Exit;
  end;

  Result := -1;
  if TypeEntries.Count <> 0 then begin
    for I := 0 to TypeEntries.Count-1 do
      if CompareText(PSetupTypeEntry(TypeEntries[I]).Name, F) = 0 then begin
        Result := I;
        Break;
      end;
  end else begin
    for I := 0 to High(DefaultTypeEntryNames) do
      if StrIComp(DefaultTypeEntryNames[I], PChar(F)) = 0 then begin
        Result := I;
        Break;
      end;
  end;
end;

function ExtractComponent(var S: String; const ComponentEntries: TList): Integer;
var
  I: Integer;
  F: String;
begin
  F := ExtractStr(S);
  if F = '' then begin
    Result := -2;
    Exit;
  end;

  Result := -1;
  for I := 0 to ComponentEntries.Count-1 do
    if CompareText(PSetupComponentEntry(ComponentEntries[I]).Name, F) = 0 then begin
      Result := I;
      Break;
    end;
end;

function ExtractTask(var S: String; const TaskEntries: TList): Integer;
var
  I: Integer;
  F: String;
begin
  F := ExtractStr(S);
  if F = '' then begin
    Result := -2;
    Exit;
  end;

  Result := -1;
  for I := 0 to TaskEntries.Count-1 do
    if CompareText(PSetupTaskEntry(TaskEntries[I]).Name, F) = 0 then begin
      Result := I;
      Break;
    end;
end;

procedure AddToCommaText(var CommaText: String; const S: String);
begin
  if CommaText <> '' then
    CommaText := CommaText + ',';
  CommaText := CommaText + S;
end;

function StrToVersionNumbers (const S: String; var VerData: TSetupVersionData): Boolean;
  procedure Split (const Str: String; var Ver: TSetupVersionDataVersion;
    var ServicePack: Word);
  var
    I, J: Integer;
    Z, B: String;
    HasBuild: Boolean;
  begin
    Cardinal(Ver) := 0;
    ServicePack := 0;
    Z := Lowercase(Str);
    I := Pos('sp', Z);
    if I <> 0 then begin
      J := StrToInt(Copy(Z, I+2, Maxint));
      if (J < Low(Byte)) or (J > High(Byte)) then
        Abort;
      ServicePack := J shl 8;
      { ^ Shift left 8 bits because we're setting the "major" service pack
        version number. This parser doesn't currently accept "minor" service
        pack version numbers. }
      SetLength (Z, I-1);
    end;
    I := Pos('.', Z);
    if I = Length(Z) then Abort;
    if I <> 0 then begin
      J := StrToInt(Copy(Z, 1, I-1));
      if (J < Low(Ver.Major)) or (J > High(Ver.Major)) then
        Abort;
      Ver.Major := J;
      Z := Copy(Z, I+1, Maxint);
      I := Pos('.', Z);
      HasBuild := I <> 0;
      if not HasBuild then
        I := Length(Z)+1;
      B := Copy(Z, I+1, Maxint);
      Z := Copy(Z, 1, I-1);
      J := StrToInt(Z);
      if (J < 0) or (J > 99) then Abort;
      if (J < 10) and (Z[1] <> '0') then J := J * 10;
      Ver.Minor := J;
      if HasBuild then begin
        J := StrToInt(B);
        if (J < Low(Ver.Build)) or (J > High(Ver.Build)) then
          Abort;
        Ver.Build := J;
      end;
    end
    else begin  { no minor version specified }
      J := StrToInt(Str);
      if (J < Low(Ver.Major)) or (J > High(Ver.Major)) then
        Abort;
      Ver.Major := J;
    end;
  end;
var
  I: Integer;
  SP: Word;
begin
  try
    I := Pos(',', S);
    if I = 0 then Abort;
    Split (Trim(Copy(S, 1, I-1)),
      TSetupVersionDataVersion(VerData.WinVersion), SP);
    if SP <> 0 then Abort;  { only NT has service packs }
    Split (Trim(Copy(S, I+1, Maxint)),
      TSetupVersionDataVersion(VerData.NTVersion), VerData.NTServicePack);
    Result := True;
  except
    Result := False;
  end;
end;

procedure TSetupCompiler.SeparateDirective (const Line: PChar;
  var Key, Value: String);
var
  P, P2: PChar;
  L: Cardinal;
begin
  Key := '';
  Value := '';
  P := Line;
  while (P^ <> #0) and (P^ <= ' ') do
    Inc (P);
  if P^ = #0 then
    Exit;
  P2 := P;
  while (P2^ <> #0) and (P2^ <> '=') do
    Inc (P2);
  L := P2 - P;
  SetLength (Key, L);
  Move (P^, Key[1], Length(Key));
  Key := TrimRight(Key);
  if P2^ = #0 then
    Exit;
  P := P2 + 1;
  while (P^ <> #0) and (P^ <= ' ') do
    Inc (P);
  if P^ = #0 then
    Exit;
  Value := TrimRight(StrPas(P));
end;

procedure TSetupCompiler.EnumSetup (const Line: PChar; const Ext: Integer);
var
  KeyName, Value: String;
  I: Integer;
  Directive: TSetupSectionDirectives;

  procedure Invalid;
  begin
    AbortCompileOnLineFmt (SCompilerEntryInvalid2, ['Setup', KeyName]);
  end;

  function StrToBool (S: String): Boolean;
  begin
    Result := False;
    S := Lowercase(S);
    if (S = '0') or (S = 'no') or (S = 'false') then
      { Result already False }
    else if (S = '1') or (S = 'yes') or (S = 'true') then
      Result := True
    else
      Invalid;
  end;

  procedure SetSetupHeaderOption (const Option: TSetupHeaderOption);
  begin
    if not StrToBool(Value) then
      Exclude (SetupHeader^.Options, Option)
    else
      Include (SetupHeader^.Options, Option);
  end;

begin
  SeparateDirective (Line, KeyName, Value);

  if KeyName = '' then
    Exit;
  I := GetEnumValue(TypeInfo(TSetupSectionDirectives), 'ss' + KeyName);
  if I = -1 then
    AbortCompileOnLineFmt (SCompilerUnknownDirective, ['Setup', KeyName]);
  Directive := TSetupSectionDirectives(I);
  if SetupDirectiveLines[Directive] <> 0 then
    AbortCompileOnLineFmt (SCompilerEntryAlreadySpecified, ['Setup', KeyName]);
  SetupDirectiveLines[Directive] := LineNumber;
  case Directive of
    ssAdminPrivilegesRequired: begin
        SetSetupHeaderOption (shAdminPrivilegesRequired);
      end;
    ssAllowNoIcons: begin
        SetSetupHeaderOption (shAllowNoIcons);
      end;
    ssAllowRootDirectory: begin
        SetSetupHeaderOption (shAllowRootDirectory);
      end;
    ssAllowUNCPath: begin
        SetSetupHeaderOption (shAllowUNCPath);
      end;
    ssAlwaysCreateUninstallIcon: begin
        SetSetupHeaderOption (shAlwaysCreateUninstallIcon);
      end;
    ssAlwaysRestart: begin
        SetSetupHeaderOption (shAlwaysRestart);
      end;
    ssAlwaysUsePersonalGroup: begin
        SetSetupHeaderOption (shAlwaysUsePersonalGroup);
      end;
    ssAlwaysShowComponentsList: begin
        SetSetupHeaderOption (shAlwaysShowComponentsList);
      end;
    ssAlwaysShowDirOnReadyPage: begin
        SetSetupHeaderOption (shAlwaysShowDirOnReadyPage);
      end;
    ssAlwaysShowGroupOnReadyPage: begin
        SetSetupHeaderOption (shAlwaysShowGroupOnReadyPage);
      end;
    ssAppCopyright: begin
        SetupHeader^.AppCopyright := Value;
      end;
    ssAppId: begin
        if Length(Value) > 127 then
          AbortCompileOnLineFmt (SCompilerEntryTooLong, ['Setup', KeyName, 127]);
        SetupHeader^.AppId := Value;
      end;
    ssAppMutex: begin
        SetupHeader^.AppMutex := Trim(Value);
      end;
    ssAppName: begin
        if Value = '' then
          Invalid;
        if Length(Value) > 127 then
          AbortCompileOnLineFmt (SCompilerEntryTooLong, ['Setup', KeyName, 127]);
        SetupHeader^.AppName := Value;
      end;
    ssAppPublisher: begin
        if Value = '' then
          Invalid;
        SetupHeader^.AppPublisher := Value;
      end;
    ssAppPublisherURL: begin
        if Value = '' then
          Invalid;
        SetupHeader^.AppPublisherURL := Value;
      end;
    ssAppSupportURL: begin
        if Value = '' then
          Invalid;
        SetupHeader^.AppSupportURL := Value;
      end;
    ssAppUpdatesURL: begin
        if Value = '' then
          Invalid;
        SetupHeader^.AppUpdatesURL := Value;
      end;
    ssAppVerName: begin
        if Value = '' then
          Invalid;
        SetupHeader^.AppVerName := Value;
      end;
    ssAppVersion: begin
        if Value = '' then
          Invalid;
        SetupHeader^.AppVersion := Value;
      end;
    ssBackColor: begin
        try
          SetupHeader^.BackColor := StringToColor(Value);
        except
          Invalid;
        end;
      end;
    ssBackColor2: begin
        try
          SetupHeader^.BackColor2 := StringToColor(Value);
        except
          Invalid;
        end;
      end;
    ssBackColorDirection: begin
        if CompareText(Value, 'toptobottom') = 0 then
          Exclude (SetupHeader^.Options, shBackColorHorizontal)
        else if CompareText(Value, 'lefttoright') = 0 then
          Include (SetupHeader^.Options, shBackColorHorizontal)
        else
          Invalid;
      end;
    ssBackSolid: begin
        BackSolid := StrToBool(Value);
      end;
    ssBits: begin  { obsolete }
        I := StrToIntDef(Value, -1);
        if (I <> 16) and (I <> 32) then
          AbortCompileOnLine (SCompilerBitsNot16or32_2);
        if I = 16 then
          AbortCompileOnLine (SCompilerNeedToUse16);
        WarningsList.Add (Format(SCompilerEntryObsolete, ['Setup', KeyName]));
      end;
    ssChangesAssociations: begin
        SetSetupHeaderOption (shChangesAssociations)
      end;
    ssCompression: begin
        Value := Lowercase(Trim(Value));
        if Value = 'none' then
          CompressLevel := 0
        else if Value = 'zip' then
          { do nothing; it's the default }
        else if Value = 'bzip' then begin
          CompressWithBzip := True;
          CompressLevel := 9;
        end
        else if Copy(Value, 1, 4) = 'zip/' then begin
          I := StrToIntDef(Copy(Value, 5, Maxint), -1);
          if (I < 1) or (I > 9) then
            Invalid;
          CompressLevel := I;
        end
        else if Copy(Value, 1, 5) = 'bzip/' then begin
          I := StrToIntDef(Copy(Value, 6, Maxint), -1);
          if (I < 1) or (I > 9) then
            Invalid;
          CompressWithBzip := True;
          CompressLevel := I;
        end
        else
          Invalid;
      end;
    ssCompressLevel: begin
        if not CompressWithBzip then begin
          I := StrToIntDef(Value, -1);
          if (I < 0) or (I > 9) then
            Invalid;
          CompressLevel := I;
        end;
        WarningsList.Add (Format(SCompilerEntrySuperseded2, ['Setup', KeyName,
           'Compression']));
      end;
    ssCreateAppDir: begin
        SetSetupHeaderOption (shCreateAppDir);
      end;
    ssCreateUninstallRegKey: begin
        SetSetupHeaderOption (shCreateUninstallRegKey);
      end;
    ssDefaultDirName: begin
        SetupHeader^.DefaultDirName := Value;
      end;
    ssDefaultGroupName: begin
        SetupHeader^.DefaultGroupName := Value;
      end;
    ssDirExistsWarning: begin
        if CompareText(Value, 'auto') = 0 then
          SetupHeader^.DirExistsWarning := ddAuto
        else if StrToBool(Value) then
          { ^ exception will be raised if Value is invalid }
          SetupHeader^.DirExistsWarning := ddYes
        else
          SetupHeader^.DirExistsWarning := ddNo;
      end;
    ssDisableAppendDir: begin
        SetSetupHeaderOption (shDisableAppendDir);
      end;
    ssDisableDirExistsWarning: begin  { obsolete; superceded by "DirExistsWarning" }
        if SetupDirectiveLines[ssDirExistsWarning] = 0 then begin
          if StrToBool(Value) then
            SetupHeader^.DirExistsWarning := ddNo
          else
            SetupHeader^.DirExistsWarning := ddAuto;
        end;
        WarningsList.Add (Format(SCompilerEntrySuperseded2, ['Setup', KeyName,
           'DirExistsWarning']));
      end;
    ssDisableDirPage: begin
        SetSetupHeaderOption (shDisableDirPage);
      end;
    ssDisableFinishedPage: begin
        SetSetupHeaderOption (shDisableFinishedPage);
      end;
    ssDisableProgramGroupPage: begin
        SetSetupHeaderOption (shDisableProgramGroupPage);
      end;
    ssDisableReadyMemo: begin
        SetSetupHeaderOption (shDisableReadyMemo);
      end;
    ssDisableReadyPage: begin
        SetSetupHeaderOption (shDisableReadyPage);
      end;
    ssDisableStartupPrompt: begin
        SetSetupHeaderOption (shDisableStartupPrompt);
      end;
    ssDiskClusterSize: begin
        Val (Value, DiskClusterSize, I);
        if I <> 0 then
          Invalid;
        if (DiskClusterSize < 1) or (DiskClusterSize > 32768) then
          AbortCompileOnLine (SCompilerDiskClusterSizeInvalid);
      end;
    ssDiskSize: begin
        Val (Value, DiskTotalBytes, I);
        if I <> 0 then
          Invalid;
        if (DiskTotalBytes < 262144) or (DiskTotalBytes > MaxDiskSize) then
          AbortCompileFmt (SCompilerDiskSizeInvalid, [262144, MaxDiskSize]);
      end;
    ssDiskSpanning: begin
        DiskSpanning := StrToBool(Value);
      end;
    ssDontMergeDuplicateFiles: begin
        DontMergeDuplicateFiles := StrToBool(Value);
      end;
    ssEnableDirDoesntExistWarning: begin
        SetSetupHeaderOption (shEnableDirDoesntExistWarning);
      end;
    ssExtraDiskSpaceRequired: begin
        Val (Value, SetupHeader^.ExtraDiskSpaceRequired, I);
        if (I <> 0) or (SetupHeader^.ExtraDiskSpaceRequired < 0) then
          Invalid;
      end;
    ssFlatComponentsList: begin
        SetSetupHeaderOption (shFlatComponentsList);
      end;
    ssInfoBeforeFile: begin
        InfoBeforeFile := RemoveQuotes(Value);
      end;
    ssInfoAfterFile: begin
        InfoAfterFile := RemoveQuotes(Value);
      end;
    ssInstallMode: begin
{
        if CompareText(Value, 'normal') = 0 then
          SetupHeader^.InstallMode := imNormal
        else if CompareText(Value, 'silent') = 0 then
          SetupHeader^.InstallMode := imSilent
        else if CompareText(Value, 'verysilent') = 0 then
          SetupHeader^.InstallMode := imVerySilent
        else
          Invalid;
}
        WarningsList.Add (Format(SCompilerInstallModeObsolete, ['Setup', KeyName]));
      end;
    ssInternalCompressLevel: begin
        I := StrToIntDef(Value, -1);
        if (I < 0) or (I > 9) then
          Invalid;
        InternalCompressLevel := I;
      end;
    ssLicenseFile: begin
        LicenseFile := RemoveQuotes(Value);
      end;
    ssMessagesFile: begin
        Value := RemoveQuotes(Value);
        if Value = '' then
          Invalid;
        MessagesFile := Value;
      end;
    ssMinVersion: begin
        if not StrToVersionNumbers(Value, SetupHeader^.MinVersion) then
          Invalid;
        if (SetupHeader^.MinVersion.WinVersion <> 0) and
           (SetupHeader^.MinVersion.WinVersion < $04000000{4.0}) then
          AbortCompileOnLineFmt (SCompilerMinVersionWinTooLow, ['4.0']);
        if (SetupHeader^.MinVersion.NTVersion <> 0) and
           (SetupHeader^.MinVersion.NTVersion < $03330000{3.51}) then
          AbortCompileOnLineFmt (SCompilerMinVersionNTTooLow, ['3.51']);
      end;
    ssOnlyBelowVersion: begin
        if not StrToVersionNumbers(Value, SetupHeader^.OnlyBelowVersion) then
          Invalid;
      end;
    ssOutputBaseFilename: begin
        Value := RemoveQuotes(Value);
        if Value = '' then
          Invalid;
        OutputBaseFilename := Value;
      end;
    ssOutputDir: begin
        Value := RemoveQuotes(Value);
        if Value = '' then
          Invalid;
        OutputDir := Value;
      end;
    ssOverwriteUninstRegEntries: begin  { obsolete; ignored }
        { was: SetSetupHeaderOption (shOverwriteUninstRegEntries); }
        WarningsList.Add (Format(SCompilerEntryObsolete, ['Setup', KeyName]));
      end;
    ssPassword: begin
        if Value <> '' then begin
          SetupHeader^.Password := GetCRC32(Value[1], Length(Value));
          Include (SetupHeader^.Options, shPassword);
        end;
      end;
    ssReserveBytes: begin
        Val (Value, ReserveBytes, I);
        if (I <> 0) or (ReserveBytes < 0) then
          Invalid;
      end;
    ssShowComponentSizes: begin
        SetSetupHeaderOption (shShowComponentSizes);
      end;
    ssSourceDir: begin
        if Value = '' then
          Invalid;
        if IsRelativePath(Value) then
          SourceDir := AddBackslash(SourceDir) + Value
        else
          SourceDir := Value;
      end;
    ssUpdateUninstallLogAppName: begin
        SetSetupHeaderOption (shUpdateUninstallLogAppName);
      end;
    ssUninstallable: begin
        SetSetupHeaderOption (shUninstallable);
      end;
    ssUninstallDisplayIcon: begin
        SetupHeader^.UninstallDisplayIcon := RemoveQuotes(Value);
      end;
    ssUninstallDisplayName: begin
        SetupHeader^.UninstallDisplayName := Value;
      end;
    ssUninstallFilesDir: begin
        Value := RemoveQuotes(Value);
        if Value = '' then
          Invalid;
        SetupHeader^.UninstallFilesDir := Value;
      end;
    ssUninstallIconName: begin
        SetupHeader^.UninstallIconName := RemoveQuotes(Value);
      end;
    ssUninstallLogMode: begin
        if CompareText(Value, 'append') = 0 then
          SetupHeader^.UninstallLogMode := lmAppend
        else if CompareText(Value, 'new') = 0 then
          SetupHeader^.UninstallLogMode := lmNew
        else if CompareText(Value, 'overwrite') = 0 then
          SetupHeader^.UninstallLogMode := lmOverwrite
        else
          Invalid;
      end;
    ssUninstallStyle: begin
        if CompareText(Value, 'classic') = 0 then
          SetupHeader.UninstallStyle := usClassic
        else if CompareText(Value, 'modern') = 0 then
          SetupHeader.UninstallStyle := usModern
        else
          Invalid;
      end;
    ssUsePreviousAppDir: begin
        SetSetupHeaderOption (shUsePreviousAppDir);
      end;
    ssUsePreviousGroup: begin
        SetSetupHeaderOption (shUsePreviousGroup);
      end;
    ssUsePreviousSetupType: begin
        SetSetupHeaderOption (shUsePreviousSetupType);
      end;
    ssUsePreviousTasks: begin
        SetSetupHeaderOption (shUsePreviousTasks);
      end;
    ssUseSetupLdr: begin
        UseSetupLdr := StrToBool(Value);
      end;
    ssWindowResizable: begin
        SetSetupHeaderOption (shWindowResizable);
      end;
    ssWindowShowCaption: begin
        SetSetupHeaderOption (shWindowShowCaption);
      end;
    ssWindowStartMaximized: begin
        SetSetupHeaderOption (shWindowStartMaximized);
      end;
    ssWindowVisible: begin
        SetSetupHeaderOption (shWindowVisible);
      end;
    ssWizardImageBackColor: begin
        try
          SetupHeader^.WizardImageBackColor := StringToColor(Value);
        except
          Invalid;
        end;
      end;
    ssWizardSmallImageBackColor: begin
        try
          SetupHeader^.WizardSmallImageBackColor := StringToColor(Value);
        except
          Invalid;
        end;
      end;
    ssWizardImageFile: begin
        Value := RemoveQuotes(Value);
        if Value = '' then
          Invalid;
        WizardImageFile := Value;
      end;
    ssWizardSmallImageFile: begin
        Value := RemoveQuotes(Value);
        if Value = '' then
          Invalid;
        WizardSmallImageFile := Value;
      end;
    ssWizardStyle: begin
        if CompareText(Value, 'classic') = 0 then begin
          if SetupDirectiveLines[ssWindowVisible] = 0 then
            Include (SetupHeader.Options, shWindowVisible);
          if SetupDirectiveLines[ssFlatComponentsList] = 0 then
            Exclude (SetupHeader.Options, shFlatComponentsList);
          if SetupDirectiveLines[ssWizardImageBackColor] = 0 then
            SetupHeader.WizardImageBackColor := DefaultWizardImageBackColorClassic;
          if SetupDirectiveLines[ssUninstallStyle] = 0 then
            SetupHeader.UninstallStyle := DefaultUninstallStyleClassic;
          if SetupDirectiveLines[ssWizardImageFile] = 0 then
            WizardImageFile := DefaultWizardImageFileClassic;
          WizardStyle := wsClassic;
        end else if CompareText(Value, 'modern') = 0 then begin
          { no-op }
        end else
          Invalid;
      end;
  end;
end;

procedure TSetupCompiler.EnumLangOptions (const Line: PChar; const Ext: Integer);
var
  KeyName, Value: String;
  I: Integer;
  Directive: TLangOptionsSectionDirectives;

  procedure Invalid;
  begin
    AbortCompileOnLineFmt (SCompilerEntryInvalid2, ['LangOptions', KeyName]);
  end;

  function TrimAndCheckName (const N: String): String;
  begin
    Result := Trim(N);
    if Result = '' then
      Invalid;
  end;

  function StrToIntCheck (const S: String): Integer;
  var
    E: Integer;
  begin
    Val (S, Result, E);
    if E <> 0 then
      Invalid;
  end;

begin
  SeparateDirective (Line, KeyName, Value);

  if KeyName = '' then
    Exit;
  I := GetEnumValue(TypeInfo(TLangOptionsSectionDirectives), 'ls' + KeyName);
  if I = -1 then
    AbortCompileOnLineFmt (SCompilerUnknownDirective, ['LangOptions', KeyName]);
  Directive := TLangOptionsSectionDirectives(I);
  if Directive in LangOptionsDirectivesSpecified then
    AbortCompileOnLineFmt (SCompilerEntryAlreadySpecified, ['LangOptions', KeyName]);
  Include (LangOptionsDirectivesSpecified, Directive);
  case Directive of
    lsCopyrightFontName: begin
        LangOptions.CopyrightFontName := TrimAndCheckName(Value);
      end;
    lsCopyrightFontSize: begin
        LangOptions.CopyrightFontSize := StrToIntCheck(Value);
      end;
    lsDialogFontName: begin
        LangOptions.DialogFontName := TrimAndCheckName(Value);
      end;
    lsDialogFontSize: begin
        LangOptions.DialogFontSize := StrToIntCheck(Value);
      end;
    lsDialogFontStandardHeight: begin
        LangOptions.DialogFontStandardHeight := StrToIntCheck(Value);
      end;
    lsLanguageID: begin
        LangOptions.LanguageID := StrToIntCheck(Value);
      end;
    lsLanguageName: begin
        LangOptions.LanguageName := TrimAndCheckName(Value);
      end;
    lsTitleFontName: begin
        LangOptions.TitleFontName := TrimAndCheckName(Value);
      end;
    lsTitleFontSize: begin
        LangOptions.TitleFontSize := StrToIntCheck(Value);
      end;
    lsWelcomeFontName: begin
        LangOptions.WelcomeFontName := TrimAndCheckName(Value);
      end;
    lsWelcomeFontSize: begin
        LangOptions.WelcomeFontSize := StrToIntCheck(Value);
      end;
  end;
end;

procedure TSetupCompiler.EnumTypes (const Line: PChar; const Ext: Integer);
const
  ParamTypesName = 'Name';
  ParamTypesDescription = 'Description';
  ParamNames: array[0..4] of TParamInfo = (
    (Name: ParamCommonFlags; Flags: []),
    (Name: ParamTypesName; Flags: [piNoEmpty]),
    (Name: ParamTypesDescription; Flags: [piNoEmpty]),
    (Name: ParamCommonMinVersion; Flags: []),
    (Name: ParamCommonOnlyBelowVersion; Flags: []));
  Flags: array[0..0] of PChar = (
    'iscustom');
var
  Params: TBreakStringArray;
  ParamNameFound: array[Low(ParamNames)..High(ParamNames)] of Boolean;
  NewTypeEntry: PSetupTypeEntry;
  P: Integer;
begin
  FillChar (ParamNameFound, SizeOf(ParamNameFound), 0);
  BreakString (Line, Params);

  NewTypeEntry := AllocMem(SizeOf(TSetupTypeEntry));
  try
    with NewTypeEntry^ do begin
      MinVersion := SetupHeader^.MinVersion;

      for P := Low(Params) to High(Params) do
        with Params[P] do begin
          if ParamName = '' then Break;
          case CompareParamName(Params[P], ParamNames, ParamNameFound) of
            0: while True do
                 case ExtractFlag(ParamData, Flags) of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownFlag2, ParamCommonFlags);
                   0: Include (Options, toIsCustom);
                 end;
            1: Name := LowerCase(ParamData);
            2: Description := ParamData;
            3: if not StrToVersionNumbers(ParamData, MinVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonMinVersion);
            4: if not StrToVersionNumbers(ParamData, OnlyBelowVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonOnlyBelowVersion);
          end;
        end;

      if not ParamNameFound[1] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamTypesName);
      if not ParamNameFound[2] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamTypesDescription);

      CheckDirConst (Description, MinVersion, []);
    end;
  except
    SEFreeRec (NewTypeEntry, SetupTypeEntryStrings);
    raise;
  end;
  TypeEntries.Add (NewTypeEntry);
end;

procedure TSetupCompiler.EnumComponents (const Line: PChar; const Ext: Integer);
const
  ParamComponentsName = 'Name';
  ParamComponentsDescription = 'Description';
  ParamComponentsExtraDiskSpaceRequired = 'ExtraDiskSpaceRequired';
  ParamComponentsTypes = 'Types';
  ParamNames: array[0..6] of TParamInfo = (
    (Name: ParamCommonFlags; Flags: []),
    (Name: ParamComponentsName; Flags: [piNoEmpty]),
    (Name: ParamComponentsDescription; Flags: [piNoEmpty]),
    (Name: ParamComponentsExtraDiskSpaceRequired; Flags: []),
    (Name: ParamComponentsTypes; Flags: []),
    (Name: ParamCommonMinVersion; Flags: []),
    (Name: ParamCommonOnlyBelowVersion; Flags: []));
  Flags: array[0..2] of PChar = (
    'fixed', 'restart', 'disablenouninstallwarning');
var
  Params: TBreakStringArray;
  ParamNameFound: array[Low(ParamNames)..High(ParamNames)] of Boolean;
  NewComponentEntry: PSetupComponentEntry;
  P, I: Integer;
begin
  FillChar (ParamNameFound, SizeOf(ParamNameFound), 0);
  BreakString (Line, Params);

  NewComponentEntry := AllocMem(SizeOf(TSetupComponentEntry));
  try
    with NewComponentEntry^ do begin
      MinVersion := SetupHeader^.MinVersion;

      for P := Low(Params) to High(Params) do
        with Params[P] do begin
          if ParamName = '' then Break;
          case CompareParamName(Params[P], ParamNames, ParamNameFound) of
            0: while True do
                 case ExtractFlag(ParamData, Flags) of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownFlag2, ParamCommonFlags);
                   0: Include (Options, coFixed);
                   1: Include (Options, coRestart);
                   2: Include (Options, coDisableNoUninstallWarning);
                 end;
            1: Name := LowerCase(ParamData);
            2: Description := ParamData;
            3: begin
               Val (ParamData, ExtraDiskSpaceRequired, I);
               if (I <> 0) or (ExtraDiskSpaceRequired < 0) then
                AbortCompileOnLine (SCompilerComponentsExtraDiskSpaceRequiredInvalid);
               end;
            4: while True do begin
                 I := ExtractType(ParamData, TypeEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownType, ParamComponentsTypes);
                   else begin
                     if TypeEntries.Count <> 0 then
                       AddToCommaText(Types, PSetupTypeEntry(TypeEntries[I]).Name)
                     else
                       AddToCommaText(Types, DefaultTypeEntryNames[I]);
                   end;
                 end;
               end;
            5: if not StrToVersionNumbers(ParamData, MinVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonMinVersion);
            6: if not StrToVersionNumbers(ParamData, OnlyBelowVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonOnlyBelowVersion);
          end;
        end;

      if not ParamNameFound[1] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamComponentsName);
      if not ParamNameFound[2] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamComponentsDescription);

      CheckDirConst (Description, MinVersion, []);

      if coRestart in Options then
        HasRestart := True;
    end;
  except
    SEFreeRec (NewComponentEntry, SetupComponentEntryStrings);
    raise;
  end;
  ComponentEntries.Add (NewComponentEntry);
end;

procedure TSetupCompiler.EnumTasks (const Line: PChar; const Ext: Integer);
const
  ParamTasksName = 'Name';
  ParamTasksDescription = 'Description';
  ParamTasksGroupDescription = 'GroupDescription';
  ParamNames: array[0..6] of TParamInfo = (
    (Name: ParamCommonFlags; Flags: []),
    (Name: ParamTasksName; Flags: [piNoEmpty, piNoQuotes]),
    (Name: ParamTasksDescription; Flags: [piNoEmpty]),
    (Name: ParamTasksGroupDescription; Flags: [piNoEmpty]),
    (Name: ParamCommonComponents; Flags: []),
    (Name: ParamCommonMinVersion; Flags: []),
    (Name: ParamCommonOnlyBelowVersion; Flags: []));
  Flags: array[0..3] of PChar = (
    'exclusive', 'unchecked', 'restart', 'checkedonce');
var
  Params: TBreakStringArray;
  ParamNameFound: array[Low(ParamNames)..High(ParamNames)] of Boolean;
  NewTaskEntry: PSetupTaskEntry;
  P, I: Integer;
begin
  FillChar (ParamNameFound, SizeOf(ParamNameFound), 0);
  BreakString (Line, Params);

  NewTaskEntry := AllocMem(SizeOf(TSetupTaskEntry));
  try
    with NewTaskEntry^ do begin
      MinVersion := SetupHeader^.MinVersion;

      for P := Low(Params) to High(Params) do
        with Params[P] do begin
          if ParamName = '' then Break;
          case CompareParamName(Params[P], ParamNames, ParamNameFound) of
            0: while True do
                 case ExtractFlag(ParamData, Flags) of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownFlag2, ParamCommonFlags);
                   0: Include (Options, toExclusive);
                   1: Include (Options, toUnchecked);
                   2: Include (Options, toRestart);
                   3: Include (Options, toCheckedOnce);
                 end;
            1: Name := ParamData;
            2: Description := ParamData;
            3: GroupDescription := ParamData;
            4: while True do begin
                 I := ExtractComponent(ParamData, ComponentEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownComponent, ParamCommonComponents);
                   else
                     AddToCommaText(Components, PSetupComponentEntry(ComponentEntries[I]).Name)
                 end;
               end;
            5: if not StrToVersionNumbers(ParamData, MinVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonMinVersion);
            6: if not StrToVersionNumbers(ParamData, OnlyBelowVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonOnlyBelowVersion);
          end;
        end;

      if not ParamNameFound[1] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamTasksName);
      if not ParamNameFound[2] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamTasksDescription);

      if (toUnchecked in Options) and (toCheckedOnce in Options) then
        AbortCompileOnLineFmt (SCompilerParamErrorBadCombo2,
          [ParamCommonFlags, 'unchecked', 'checkedonce']);

      CheckDirConst (Description, MinVersion, []);
      CheckDirConst (GroupDescription, MinVersion, []);

      if toRestart in Options then
        HasRestart := True;
    end;
  except
    SEFreeRec (NewTaskEntry, SetupTaskEntryStrings);
    raise;
  end;
  TaskEntries.Add (NewTaskEntry);
end;

procedure TSetupCompiler.EnumDirs (const Line: PChar; const Ext: Integer);
const
  ParamDirsName = 'Name';
  ParamDirsAttribs = 'Attribs';
  ParamNames: array[0..6] of TParamInfo = (
    (Name: ParamCommonFlags; Flags: []),
    (Name: ParamDirsName; Flags: [piNoEmpty, piNoQuotes]),
    (Name: ParamCommonComponents; Flags: []),
    (Name: ParamCommonTasks; Flags: []),
    (Name: ParamCommonMinVersion; Flags: []),
    (Name: ParamCommonOnlyBelowVersion; Flags: []),
    (Name: ParamDirsAttribs; Flags: []));
  Flags: array[0..2] of PChar = (
    'uninsneveruninstall', 'deleteafterinstall', 'uninsalwaysuninstall');
  AttribsFlags: array[0..2] of PChar = (
    'readonly', 'hidden', 'system');
var
  Params: TBreakStringArray;
  ParamNameFound: array[Low(ParamNames)..High(ParamNames)] of Boolean;
  NewDirEntry: PSetupDirEntry;
  P, I: Integer;
begin
  FillChar (ParamNameFound, SizeOf(ParamNameFound), 0);
  BreakString (Line, Params);

  NewDirEntry := AllocMem(SizeOf(TSetupDirEntry));
  try
    with NewDirEntry^ do begin
      MinVersion := SetupHeader^.MinVersion;

      for P := Low(Params) to High(Params) do
        with Params[P] do begin
          if ParamName = '' then Break;
          case CompareParamName(Params[P], ParamNames, ParamNameFound) of
            0: while True do
                 case ExtractFlag(ParamData, Flags) of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownFlag2, ParamCommonFlags);
                   0: Include (Options, doUninsNeverUninstall);
                   1: Include (Options, doDeleteAfterInstall);
                   2: Include (Options, doUninsAlwaysUninstall);
                 end;
            1: DirName := ParamData;
            2: while True do begin
                 I := ExtractComponent(ParamData, ComponentEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownComponent, ParamCommonComponents);
                   else
                     AddToCommaText(Components, PSetupComponentEntry(ComponentEntries[I]).Name)
                 end;
               end;
            3: while True do begin
                 I := ExtractTask(ParamData, TaskEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownTask, ParamCommonTasks);
                   else
                     AddToCommaText(Tasks, PSetupTaskEntry(TaskEntries[I]).Name)
                 end;
               end;
            4: if not StrToVersionNumbers(ParamData, MinVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonMinVersion);
            5: if not StrToVersionNumbers(ParamData, OnlyBelowVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonOnlyBelowVersion);
            6: while True do
                 case ExtractFlag(ParamData, AttribsFlags) of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownFlag2, ParamDirsAttribs);
                   0: Attribs := Attribs or faReadOnly;
                   1: Attribs := Attribs or faHidden;
                   2: Attribs := Attribs or faSysFile;
                 end;
          end;
        end;

      if not ParamNameFound[1] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamDirsName);

      if (doUninsNeverUninstall in Options) and
         (doUninsAlwaysUninstall in Options) then
        AbortCompileOnLineFmt (SCompilerParamErrorBadCombo2,
          [ParamCommonFlags, 'uninsneveruninstall', 'uninsalwaysuninstall']);

      CheckDirConst (DirName, MinVersion, []);
    end;
  except
    SEFreeRec (NewDirEntry, SetupDirEntryStrings);
    raise;
  end;
  DirEntries.Add (NewDirEntry);
end;

const
  BadPMChars = ',()[]';

function ContainsBadPMChars (const S: String): Boolean;
var
  I: Integer;
begin
  for I := 1 to Length(S) do
    if Pos(S[I], BadPMChars) <> 0 then begin
      Result := True;
      Exit;
    end;
  Result := False;
end;

function SpaceString (const S: String): String;
var
  I: Integer;
begin
  Result := '';
  for I := 1 to Length(S) do begin
    if S[I] = ' ' then Continue;
    if Result <> '' then Result := Result + ' ';
    Result := Result + S[I];
  end;
end;

type
  TMenuKeyCap = (mkcBkSp, mkcTab, mkcEsc, mkcEnter, mkcSpace, mkcPgUp,
    mkcPgDn, mkcEnd, mkcHome, mkcLeft, mkcUp, mkcRight, mkcDown, mkcIns,
    mkcDel, mkcShift, mkcCtrl, mkcAlt);

{$IFDEF Delphi3OrHigher}
var
  MenuKeyCaps: array[TMenuKeyCap] of string = (
    SmkcBkSp, SmkcTab, SmkcEsc, SmkcEnter, SmkcSpace, SmkcPgUp,
    SmkcPgDn, SmkcEnd, SmkcHome, SmkcLeft, SmkcUp, SmkcRight,
    SmkcDown, SmkcIns, SmkcDel, SmkcShift, SmkcCtrl, SmkcAlt);
{$ELSE}
var
  MenuKeyCaps: array[TMenuKeyCap] of string;

const
  MenuKeyCapIDs: array[TMenuKeyCap] of Word = (
    SmkcBkSp, SmkcTab, SmkcEsc, SmkcEnter, SmkcSpace, SmkcPgUp,
    SmkcPgDn, SmkcEnd, SmkcHome, SmkcLeft, SmkcUp, SmkcRight,
    SmkcDown, SmkcIns, SmkcDel, SmkcShift, SmkcCtrl, SmkcAlt);
{$ENDIF}

procedure TSetupCompiler.EnumIcons (const Line: PChar; const Ext: Integer);

  {$IFNDEF Delphi3OrHigher}
  procedure LoadStrings;
  var
    I: TMenuKeyCap;
  begin
    for I := Low(TMenuKeyCap) to High(TMenuKeyCap) do
      MenuKeyCaps[I] := LoadStr(MenuKeyCapIDs[I]);
  end;
  {$ENDIF}

  function HotKeyToText(HotKey: Word): string;

    function GetSpecialName(HotKey: Word): string;
    var
      ScanCode: Integer;
      KeyName: array[0..255] of Char;
    begin
      Result := '';
      ScanCode := MapVirtualKey(WordRec(HotKey).Lo, 0) shl 16;
      if ScanCode <> 0 then
      begin
        GetKeyNameText(ScanCode, KeyName, SizeOf(KeyName));
        if (KeyName[1] = #0) and (KeyName[0] <> #0) then
          GetSpecialName := KeyName;
      end;
    end;

  var
    Name: string;
  begin
    case WordRec(HotKey).Lo of
      $08, $09:
        Name := MenuKeyCaps[TMenuKeyCap(Ord(mkcBkSp) + WordRec(HotKey).Lo - $08)];
      $0D: Name := MenuKeyCaps[mkcEnter];
      $1B: Name := MenuKeyCaps[mkcEsc];
      $20..$28:
        Name := MenuKeyCaps[TMenuKeyCap(Ord(mkcSpace) + WordRec(HotKey).Lo - $20)];
      $2D..$2E:
        Name := MenuKeyCaps[TMenuKeyCap(Ord(mkcIns) + WordRec(HotKey).Lo - $2D)];
      $30..$39: Name := Chr(WordRec(HotKey).Lo - $30 + Ord('0'));
      $41..$5A: Name := Chr(WordRec(HotKey).Lo - $41 + Ord('A'));
      $60..$69: Name := Chr(WordRec(HotKey).Lo - $60 + Ord('0'));
      $70..$87: Name := 'F' + IntToStr(WordRec(HotKey).Lo - $6F);
    else
      Name := GetSpecialName(HotKey);
    end;
    if Name <> '' then
    begin
      Result := '';
      if HotKey and (HOTKEYF_SHIFT shl 8) <> 0 then Result := Result + MenuKeyCaps[mkcShift];
      if HotKey and (HOTKEYF_CONTROL shl 8) <> 0 then Result := Result + MenuKeyCaps[mkcCtrl];
      if HotKey and (HOTKEYF_ALT shl 8) <> 0 then Result := Result + MenuKeyCaps[mkcAlt];
      Result := Result + Name;
    end
    else Result := '';
  end;

  function TextToHotKey(Text: string): Word;

    function CompareFront(var Text: string; const Front: string): Boolean;
    begin
      Result := False;
      if CompareText(Copy(Text, 1, Length(Front)), Front) = 0 then
      begin
        Result := True;
        Delete(Text, 1, Length(Front));
      end;
    end;

  var
    Key: Word;
    Shift: Word;
  begin
    Result := 0;
    Shift := 0;
    while True do
    begin
      if CompareFront(Text, MenuKeyCaps[mkcShift]) then Shift := Shift or HOTKEYF_SHIFT
      else if CompareFront(Text, '^') then Shift := Shift or HOTKEYF_CONTROL
      else if CompareFront(Text, MenuKeyCaps[mkcCtrl]) then Shift := Shift or HOTKEYF_CONTROL
      else if CompareFront(Text, MenuKeyCaps[mkcAlt]) then Shift := Shift or HOTKEYF_ALT
      else Break;
    end;
    if Text = '' then Exit;
    for Key := $08 to $255 do { Copy range from table in HotKeyToText }
      if AnsiCompareText(Text, HotKeyToText(Key)) = 0 then
      begin
        Result := Key or (Shift shl 8);
        Exit;
      end;
  end;

  function GenerateTaskName: String;
  var
    I, N: Integer;
    S: String;
    Found: Boolean;
  begin
    N := 0;
    repeat
      S := IntToStr(N);
      Inc(N);
      Found := False;
      for I := 0 to TaskEntries.Count-1 do begin
        if PSetupTaskEntry(TaskEntries[i]).Name = S then begin
          Found := True;
          Break;
        end;
      end;
    until not Found;
    Result := S;
  end;

const
  ParamIconsName = 'Name';
  ParamIconsFilename = 'Filename';
  ParamIconsParameters = 'Parameters';
  ParamIconsWorkingDir = 'WorkingDir';
  ParamIconsHotKey = 'HotKey';
  ParamIconsIconFilename = 'IconFilename';
  ParamIconsIconIndex = 'IconIndex';
  ParamDescription = 'Description';
  ParamIconsComment = 'Comment';
  ParamNames: array[0..13] of TParamInfo = (
    (Name: ParamCommonFlags; Flags: []),
    (Name: ParamIconsName; Flags: [piNoEmpty, piNoQuotes]),
    (Name: ParamIconsFilename; Flags: [piNoEmpty, piNoQuotes]),
    (Name: ParamIconsParameters; Flags: []),
    (Name: ParamIconsWorkingDir; Flags: [piNoQuotes]),
    (Name: ParamIconsHotKey; Flags: []),
    (Name: ParamIconsIconFilename; Flags: [piNoQuotes]),
    (Name: ParamIconsIconIndex; Flags: []),
    (Name: ParamDescription; Flags: []),
    (Name: ParamCommonComponents; Flags: []),
    (Name: ParamCommonTasks; Flags: []),
    (Name: ParamCommonMinVersion; Flags: []),
    (Name: ParamCommonOnlyBelowVersion; Flags: []),
    (Name: ParamIconsComment; Flags: []));
  Flags: array[0..8] of PChar = (
    'uninsneveruninstall', 'runminimized', 'createonlyiffileexists',
    'useapppaths', 'closeonexit', 'dontcloseonexit', 'runmaximized',
    'showcheckbox', 'unchecked');
var
  Params: TBreakStringArray;
  ParamNameFound: array[Low(ParamNames)..High(ParamNames)] of Boolean;
  NewIconEntry: PSetupIconEntry;
  NewTaskEntry: PSetupTaskEntry;
  P, I: Integer;
  Description, S: String;
  ShowCheckBox, Unchecked: Boolean;
begin
  {$IFNDEF Delphi3OrHigher}
  LoadStrings;
  {$ENDIF}

  FillChar (ParamNameFound, SizeOf(ParamNameFound), 0);
  BreakString (Line, Params);

  NewIconEntry := AllocMem(SizeOf(TSetupIconEntry));
  try
    with NewIconEntry^ do begin
      MinVersion := SetupHeader^.MinVersion;
      ShowCmd := SW_SHOWNORMAL;

      Description := '';
      ShowCheckBox := False;
      Unchecked := False;

      for P := Low(Params) to High(Params) do
        with Params[P] do begin
          if ParamName = '' then Break;
          case CompareParamName(Params[P], ParamNames, ParamNameFound) of
            0: while True do
                 case ExtractFlag(ParamData, Flags) of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownFlag2, ParamCommonFlags);
                   0: Include (Options, ioUninsNeverUninstall);
                   1: ShowCmd := SW_SHOWMINNOACTIVE;
                   2: Include (Options, ioCreateOnlyIfFileExists);
                   3: Include (Options, ioUseAppPaths);
                   4: CloseOnExit := icYes;
                   5: CloseOnExit := icNo;
                   6: ShowCmd := SW_SHOWMAXIMIZED;
                   7: ShowCheckBox := True;
                   8: Unchecked := True;
                 end;
            1: IconName := ParamData;
            2: Filename := ParamData;
            3: Parameters := ParamData;
            4: WorkingDir := ParamData;
            5: begin
                 HotKey := TextToHotKey(ParamData);
                 if HotKey = 0 then
                   AbortCompileParamError (SCompilerParamInvalid2, ParamIconsHotKey)
               end;
            6: IconFilename := ParamData;
            7: begin
                 try
                   IconIndex := StrToInt(ParamData);
                 except
                   AbortCompileOnLine (SCompilerIconsIconIndexInvalid);
                 end;
               end;
            8: Description := ParamData;
            9: while True do begin
                 I := ExtractComponent(ParamData, ComponentEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownComponent, ParamCommonComponents);
                   else
                     AddToCommaText(Components, PSetupComponentEntry(ComponentEntries[I]).Name)
                 end;
               end;
            10:while True do begin
                 I := ExtractTask(ParamData, TaskEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownTask, ParamCommonTasks);
                   else
                     AddToCommaText(Tasks, PSetupTaskEntry(TaskEntries[I]).Name)
                 end;
               end;
            11: if not StrToVersionNumbers(ParamData, MinVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonMinVersion);
            12: if not StrToVersionNumbers(ParamData, OnlyBelowVersion) then
                  AbortCompileParamError (SCompilerParamInvalid2, ParamCommonOnlyBelowVersion);
            13: Comment := ParamData;
          end;
        end;

      if not ParamNameFound[1] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamIconsName);
      if not ParamNameFound[2] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamIconsFilename);

      if ShowCheckBox then begin
        if not ParamNameFound[8] then
          AbortCompileParamError (SCompilerParamNotSpecified, ParamDescription);
        if Description = '' then
          AbortCompileParamError (SCompilerParamEmpty2, ParamDescription);
      end;

      if MinVersionLessThan4(MinVersion) then begin
        S := IconName;
        if Copy(S, 1, 8) <> '{group}\' then
          AbortCompileOnLine (SCompilerIconsNameNoFolderConsts);
        Delete (S, 1, 8);
        if (AnsiPos('\', S) <> 0) or (AnsiPos('{', S) <> 0) then
          AbortCompileOnLine (SCompilerIconsNameNoFolderConsts);
        if Pos('"', Parameters) <> 0 then
          { Win 3.x's Program Manager has problems with quotes }
          AbortCompileParamError (SCompilerParamIncludesQuotes +
            SNewLine2 + MinVersionErrorMessage(True, $400, $400),
            ParamIconsParameters);
      end;

      if Pos('"', IconName) <> 0 then
        AbortCompileParamError (SCompilerParamNoQuotes2, ParamIconsName);
      if AnsiPos('\', IconName) = 0 then
        AbortCompileOnLine (SCompilerIconsNamePathNotSpecified);

      if (IconIndex <> 0) and (IconFilename = '') then
        IconFilename := Filename;

      if MinVersionLessThan4(MinVersion) and ContainsBadPMChars(Parameters) then
        AbortCompileOnLine (SCompilerIconsParametersBadChar);
      S := IconName;
      if Copy(S, 1, 8) = '{group}\' then
        Delete (S, 1, 8);
      CheckDirConst (S, MinVersion, []);
      CheckDirConst (Filename, MinVersion, []);
      CheckDirConst (Parameters, MinVersion, []);
      CheckDirConst (WorkingDir, MinVersion, []);
      CheckDirConst (IconFilename, MinVersion, []);
      CheckDirConst (Comment, MinVersion, []);
    end;
  except
    SEFreeRec (NewIconEntry, SetupIconEntryStrings);
    raise;
  end;

  if ShowCheckBox then begin
    NewTaskEntry := AllocMem(Sizeof(TSetupTaskEntry));
    NewTaskEntry.Name := GenerateTaskName();
    NewTaskEntry.Description := Description;
    NewTaskEntry.Components := NewIconEntry.Components;
    if Unchecked then
      Include(NewTaskEntry.Options, toUnchecked);
    NewTaskEntry.MinVersion := NewIconEntry.MinVersion;
    NewTaskEntry.OnlyBelowVersion := NewIconEntry.OnlyBelowVersion;
    TaskEntries.Add(NewTaskEntry);

    NewIconEntry.Tasks := NewTaskEntry.Name;
  end;
  IconEntries.Add (NewIconEntry);
end;

procedure TSetupCompiler.EnumINI (const Line: PChar; const Ext: Integer);
const
  ParamIniFilename = 'Filename';
  ParamIniSection = 'Section';
  ParamIniKey = 'Key';
  ParamIniString = 'String';
  ParamNames: array[0..8] of TParamInfo = (
    (Name: ParamCommonFlags; Flags: []),
    (Name: ParamIniFilename; Flags: [piNoQuotes]),
    (Name: ParamIniSection; Flags: [piNoEmpty]),
    (Name: ParamIniKey; Flags: [piNoEmpty]),
    (Name: ParamIniString; Flags: []),
    (Name: ParamCommonComponents; Flags: []),
    (Name: ParamCommonTasks; Flags: []),
    (Name: ParamCommonMinVersion; Flags: []),
    (Name: ParamCommonOnlyBelowVersion; Flags: []));
  Flags: array[0..3] of PChar = (
    'uninsdeleteentry', 'uninsdeletesection', 'createkeyifdoesntexist',
    'uninsdeletesectionifempty');
var
  Params: TBreakStringArray;
  ParamNameFound: array[Low(ParamNames)..High(ParamNames)] of Boolean;
  NewIniEntry: PSetupIniEntry;
  P, I: Integer;
begin
  FillChar (ParamNameFound, SizeOf(ParamNameFound), 0);
  BreakString (Line, Params);

  NewIniEntry := AllocMem(SizeOf(TSetupIniEntry));
  try
    with NewIniEntry^ do begin
      MinVersion := SetupHeader^.MinVersion;

      for P := Low(Params) to High(Params) do
        with Params[P] do begin
          if ParamName = '' then Break;
          case CompareParamName(Params[P], ParamNames, ParamNameFound) of
            0: while True do
                 case ExtractFlag(ParamData, Flags) of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownFlag2, ParamCommonFlags);
                   0: Include (Options, ioUninsDeleteEntry);
                   1: Include (Options, ioUninsDeleteEntireSection);
                   2: Include (Options, ioCreateKeyIfDoesntExist);
                   3: Include (Options, ioUninsDeleteSectionIfEmpty);
                 end;
            1: Filename := ParamData;
            2: Section := ParamData;
            3: Entry := ParamData;
            4: begin
                 Value := ParamData;
                 Include (Options, ioHasValue);
               end;
            5: while True do begin
                 I := ExtractComponent(ParamData, ComponentEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownComponent, ParamCommonComponents);
                   else
                     AddToCommaText(Components, PSetupComponentEntry(ComponentEntries[I]).Name)
                 end;
               end;
            6: while True do begin
                 I := ExtractTask(ParamData, TaskEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownTask, ParamCommonTasks);
                   else
                     AddToCommaText(Tasks, PSetupTaskEntry(TaskEntries[I]).Name)
                 end;
               end;
            7: if not StrToVersionNumbers(ParamData, MinVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonMinVersion);
            8: if not StrToVersionNumbers(ParamData, OnlyBelowVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonOnlyBelowVersion);
          end;
        end;

      if not ParamNameFound[1] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamIniFilename);
      if not ParamNameFound[2] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamIniSection);

      if (ioUninsDeleteEntry in Options) and
         (ioUninsDeleteEntireSection in Options) then
        AbortCompileOnLineFmt (SCompilerParamErrorBadCombo2,
          [ParamCommonFlags, 'uninsdeleteentry', 'uninsdeletesection']);
      if (ioUninsDeleteEntireSection in Options) and
         (ioUninsDeleteSectionIfEmpty in Options) then
        AbortCompileOnLineFmt (SCompilerParamErrorBadCombo2,
          [ParamCommonFlags, 'uninsdeletesection', 'uninsdeletesectionifempty']);

      CheckDirConst (Filename, MinVersion, []);
      CheckDirConst (Section, MinVersion, []);
      CheckDirConst (Entry, MinVersion, []);
      CheckDirConst (Value, MinVersion, []);
    end;
  except
    SEFreeRec (NewIniEntry, SetupIniEntryStrings);
    raise;
  end;
  IniEntries.Add (NewIniEntry);
end;

procedure TSetupCompiler.EnumRegistry (const Line: PChar; const Ext: Integer);
const
  ParamRegistryRoot = 'Root';
  ParamRegistrySubkey = 'Subkey';
  ParamRegistryValueType = 'ValueType';
  ParamRegistryValueName = 'ValueName';
  ParamRegistryValueData = 'ValueData';
  ParamNames: array[0..9] of TParamInfo = (
    (Name: ParamCommonFlags; Flags: []),
    (Name: ParamRegistryRoot; Flags: []),
    (Name: ParamRegistrySubkey; Flags: []),
    (Name: ParamRegistryValueType; Flags: []),
    (Name: ParamRegistryValueName; Flags: []),
    (Name: ParamRegistryValueData; Flags: []),
    (Name: ParamCommonComponents; Flags: []),
    (Name: ParamCommonTasks; Flags: []),
    (Name: ParamCommonMinVersion; Flags: []),
    (Name: ParamCommonOnlyBelowVersion; Flags: []));
  Flags: array[0..9] of PChar = (
    'createvalueifdoesntexist', 'uninsdeletevalue', 'uninsdeletekey',
    'uninsdeletekeyifempty', 'uninsclearvalue', 'preservestringtype',
    'deletekey', 'deletevalue', 'noerror', 'dontcreatekey');

  function ConvertBinaryString (const S: String): String;
    procedure Invalid;
    begin
      AbortCompileParamError (SCompilerParamInvalid2, ParamRegistryValueData);
    end;
  var
    I: Integer;
    C: Char;
    B: Byte;
    N: Integer;
    procedure EndByte;
    begin
      case N of
        0: ;
        2: begin
             Result := Result + Chr(B);
             N := 0;
             B := 0;
           end;
      else
        Invalid;
      end;
    end;
  begin
    Result := '';
    N := 0;
    B := 0;
    for I := 1 to Length(S) do begin
      C := UpCase(S[I]);
      case C of
        ' ': EndByte;
        '0'..'9': begin
               Inc (N);
               if N > 2 then
                 Invalid;
               B := (B shl 4) or (Ord(C) - Ord('0'));
             end;
        'A'..'F': begin
               Inc (N);
               if N > 2 then
                 Invalid;
               B := (B shl 4) or (10 + Ord(C) - Ord('A'));
             end;
      else
        Invalid;
      end;
    end;
    EndByte;
  end;

  function ConvertDWordString (const S: String): String;
  var
    DW: Longint;
    E: Integer;
  begin
    Val (S, DW, E);
    if E <> 0 then
      AbortCompileParamError (SCompilerParamInvalid2, ParamRegistryValueData);
    SetLength (Result, SizeOf(Longint));
    Longint((@Result[1])^) := DW;
  end;

var
  Params: TBreakStringArray;
  ParamNameFound: array[Low(ParamNames)..High(ParamNames)] of Boolean;
  NewRegistryEntry: PSetupRegistryEntry;
  P, I: Integer;
  AData: String;
begin
  FillChar (ParamNameFound, SizeOf(ParamNameFound), 0);
  BreakString (Line, Params);

  NewRegistryEntry := AllocMem(SizeOf(TSetupRegistryEntry));
  try
    with NewRegistryEntry^ do begin
      MinVersion := SetupHeader^.MinVersion;

      for P := Low(Params) to High(Params) do
        with Params[P] do begin
          if ParamName = '' then Break;
          case CompareParamName(Params[P], ParamNames, ParamNameFound) of
            0: while True do
                 case ExtractFlag(ParamData, Flags) of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownFlag2, ParamCommonFlags);
                   0: Include (Options, roCreateValueIfDoesntExist);
                   1: Include (Options, roUninsDeleteValue);
                   2: Include (Options, roUninsDeleteEntireKey);
                   3: Include (Options, roUninsDeleteEntireKeyIfEmpty);
                   4: Include (Options, roUninsClearValue);
                   5: Include (Options, roPreserveStringType);
                   6: Include (Options, roDeleteKey);
                   7: Include (Options, roDeleteValue);
                   8: Include (Options, roNoError);
                   9: Include (Options, roDontCreateKey);
                 end;
            1: begin
                 ParamData := Uppercase(Trim(ParamData));
                 if ParamData = 'HKCR' then
                   RootKey := HKEY_CLASSES_ROOT
                 else
                 if ParamData = 'HKCU' then
                   RootKey := HKEY_CURRENT_USER
                 else
                 if ParamData = 'HKLM' then
                   RootKey := HKEY_LOCAL_MACHINE
                 else
                 if ParamData = 'HKU' then
                   RootKey := HKEY_USERS
                 else
                 if ParamData = 'HKCC' then
                   RootKey := HKEY_CURRENT_CONFIG
                 else
                   AbortCompileParamError (SCompilerParamInvalid2, ParamRegistryRoot);
               end;
            2: begin
                 if (ParamData <> '') and (ParamData[1] = '\') then
                   AbortCompileParamError (SCompilerParamNoPrecedingBackslash, ParamRegistrySubkey);
                 Subkey := RemoveBackslash(ParamData);
               end;
            3: begin
                 ParamData := Uppercase(Trim(ParamData));
                 if ParamData = 'NONE' then
                   Typ := rtNone
                 else if ParamData = 'STRING' then
                   Typ := rtString
                 else if ParamData = 'EXPANDSZ' then
                   Typ := rtExpandString
                 else if ParamData = 'MULTISZ' then
                   Typ := rtMultiString
                 else if ParamData = 'DWORD' then
                   Typ := rtDWord
                 else if ParamData = 'BINARY' then
                   Typ := rtBinary
                 else
                   AbortCompileParamError (SCompilerParamInvalid2, ParamRegistryValueType);
               end;
            4: ValueName := ParamData;
            5: AData := ParamData;
            6: while True do begin
                 I := ExtractComponent(ParamData, ComponentEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownComponent, ParamCommonComponents);
                   else
                     AddToCommaText(Components, PSetupComponentEntry(ComponentEntries[I]).Name)
                 end;
               end;
            7: while True do begin
                 I := ExtractTask(ParamData, TaskEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownTask, ParamCommonTasks);
                   else
                     AddToCommaText(Tasks, PSetupTaskEntry(TaskEntries[I]).Name)
                 end;
               end;
            8: if not StrToVersionNumbers(ParamData, MinVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonMinVersion);
            9:if not StrToVersionNumbers(ParamData, OnlyBelowVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonOnlyBelowVersion);
          end;
        end;

      if not ParamNameFound[1] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamRegistryRoot);
      if not ParamNameFound[2] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamRegistrySubkey);

      if (roUninsDeleteEntireKey in Options) and
         (roUninsDeleteEntireKeyIfEmpty in Options) then
        AbortCompileOnLineFmt (SCompilerParamErrorBadCombo2,
          [ParamCommonFlags, 'uninsdeletekey', 'uninsdeletekeyifempty']);
      if (roUninsDeleteEntireKey in Options) and
         (roUninsClearValue in Options) then
        AbortCompileOnLineFmt (SCompilerParamErrorBadCombo2,
          [ParamCommonFlags, 'uninsclearvalue', 'uninsdeletekey']);
      if (roUninsDeleteValue in Options) and
         (roUninsDeleteEntireKey in Options) then
        AbortCompileOnLineFmt (SCompilerParamErrorBadCombo2,
          [ParamCommonFlags, 'uninsdeletevalue', 'uninsdeletekey']);
      if (roUninsDeleteValue in Options) and
         (roUninsClearValue in Options) then
        AbortCompileOnLineFmt (SCompilerParamErrorBadCombo2,
          [ParamCommonFlags, 'uninsdeletevalue', 'uninsclearvalue']);
      if ((roUninsDeleteEntireKey in Options) or (roDeleteKey in Options)) and
         (RootKey <> HKEY_CLASSES_ROOT) and (AnsiPos('\', Subkey) = 0) then
        AbortCompileOnLine (SCompilerRegistryCantDeleteKey);

      case Typ of
        rtString, rtExpandString, rtMultiString:
          ValueData := AData;
        rtDWord:
          ValueData := ConvertDWordString(AData);
        rtBinary:
          ValueData := ConvertBinaryString(AData);
      end;

      CheckDirConst (Subkey, MinVersion, []);
      CheckDirConst (ValueName, MinVersion, []);
      case Typ of
        rtString, rtExpandString:
          CheckDirConst (ValueData, MinVersion, [acOldData]);
        rtMultiString:
          CheckDirConst (ValueData, MinVersion, [acOldData, acBreak]);
      end;
    end;
  except
    SEFreeRec (NewRegistryEntry, SetupRegistryEntryStrings);
    raise;
  end;
  RegistryEntries.Add (NewRegistryEntry);
end;

procedure TSetupCompiler.EnumDelete (const Line: PChar; const Ext: Integer);
const
  ParamDeleteType = 'Type';
  ParamDeleteName = 'Name';
  ParamNames: array[0..5] of TParamInfo = (
    (Name: ParamDeleteType; Flags: []),
    (Name: ParamDeleteName; Flags: [piNoEmpty]),
    (Name: ParamCommonComponents; Flags: []),
    (Name: ParamCommonTasks; Flags: []),
    (Name: ParamCommonMinVersion; Flags: []),
    (Name: ParamCommonOnlyBelowVersion; Flags: []));
  Types: array[TSetupDeleteType] of PChar = (
    'files', 'filesandordirs', 'dirifempty');
var
  Params: TBreakStringArray;
  ParamNameFound: array[Low(ParamNames)..High(ParamNames)] of Boolean;
  NewDeleteEntry: PSetupDeleteEntry;
  P, I: Integer;
  Valid: Boolean;
  J: TSetupDeleteType;
begin
  FillChar (ParamNameFound, SizeOf(ParamNameFound), 0);
  BreakString (Line, Params);

  NewDeleteEntry := AllocMem(SizeOf(TSetupDeleteEntry));
  try
    with NewDeleteEntry^ do begin
      MinVersion := SetupHeader^.MinVersion;

      for P := Low(Params) to High(Params) do
        with Params[P] do begin
          if ParamName = '' then Break;
          case CompareParamName(Params[P], ParamNames, ParamNameFound) of
            0: begin
                 ParamData := Trim(ParamData);
                 Valid := False;
                 for J := Low(J) to High(J) do
                   if StrIComp(Types[J], StringAsPChar(ParamData)) = 0 then begin
                     DeleteType := J;
                     Valid := True;
                     Break;
                   end;
                 if not Valid then
                   AbortCompileParamError (SCompilerParamInvalid2, ParamDeleteType);
               end;
            1: Name := ParamData;
            2: while True do begin
                 I := ExtractComponent(ParamData, ComponentEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownComponent, ParamCommonComponents);
                   else
                     AddToCommaText(Components, PSetupComponentEntry(ComponentEntries[I]).Name)
                 end;
               end;
            3: while True do begin
                 I := ExtractTask(ParamData, TaskEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownTask, ParamCommonTasks);
                   else
                     AddToCommaText(Tasks, PSetupTaskEntry(TaskEntries[I]).Name)
                 end;
               end;
            4: if not StrToVersionNumbers(ParamData, MinVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonMinVersion);
            5: if not StrToVersionNumbers(ParamData, OnlyBelowVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonOnlyBelowVersion);
          end;
        end;

      if not ParamNameFound[0] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamDeleteType);
      if not ParamNameFound[1] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamDeleteName);

      CheckDirConst (Name, MinVersion, []);
    end;
  except
    SEFreeRec (NewDeleteEntry, SetupDeleteEntryStrings);
    raise;
  end;
  if Ext = 0 then
    InstallDeleteEntries.Add (NewDeleteEntry)
  else
    UninstallDeleteEntries.Add (NewDeleteEntry);
end;

type
  PExtraData = ^TExtraData;
  TExtraData = record
    Compiler: TSetupCompiler;
    SourceFile, DestFile: TFileStream;
    CurDisk: Integer;
    DiskBytesLeft: Longint;
    DestFileIsDisk: Boolean;
  end;

procedure EndDisk (var Data: TExtraData);
var
  DiskHeader: TDiskHeader;
begin
  with Data do begin
    if Assigned(DestFile) then begin
      if DestFileIsDisk then begin
        DiskHeader.TotalSize := DestFile.Size;
        DestFile.Position := SizeOf(DiskID);
        DestFile.WriteBuffer (DiskHeader, SizeOf(DiskHeader));
      end;
      DestFile.Free;
      DestFile := nil;
    end;
  end;
end;

procedure NewDisk (var Data: TExtraData; const Filename: String);
var
  DiskHeader: TDiskHeader;
begin
  EndDisk (Data);
  with Data do begin
    Inc (Data.CurDisk);
    if (Data.CurDisk > 1) and not Compiler.DiskSpanning then
      Compiler.AbortCompileFmt (SCompilerMustUseDiskSpanning,
        [Compiler.DiskTotalBytes]);
    if Filename = '' then begin
      DestFileIsDisk := True;
      DestFile := TFileStream.Create(Format('%s%s-%d.bin',
        [Compiler.OutputDir, Compiler.OutputBaseFilename, CurDisk]),
        fmCreate);
      DestFile.WriteBuffer (DiskID, SizeOf(DiskID));
      DiskHeader.TotalSize := 0;
      DestFile.WriteBuffer (DiskHeader, SizeOf(DiskHeader));
      DiskBytesLeft := Compiler.DiskTotalBytes - (SizeOf(DiskID) + SizeOf(DiskHeader));
    end
    else begin
      DestFileIsDisk := False;
      DestFile := TFileStream.Create(Filename, fmOpenReadWrite or fmShareExclusive);
      DestFile.Seek (0, soFromEnd);
      DiskBytesLeft := Compiler.DiskTotalBytes - DestFile.Position;
    end;
  end;
end;

function ReadProc (var Buf; MaxBytes: Cardinal; ExtraData: Longint): Cardinal; far;
var
  Bytes: Cardinal;
begin
  with PExtraData(ExtraData)^ do begin
    Bytes := SourceFile.Size - SourceFile.Position;
    if Bytes > MaxBytes then Bytes := MaxBytes;
    SourceFile.ReadBuffer (Buf, Bytes);
    Result := Bytes;
  end;
end;

function WriteProc (var Buf; BufSize: Cardinal; ExtraData: Longint): Cardinal; far;
var
  P: Pointer;
  S: Cardinal;
begin
  Result := BufSize;
  with PExtraData(ExtraData)^ do begin
    Compiler.CallIdleProc;
    P := @Buf;
    while BufSize > 0 do begin
      S := BufSize;
      if DiskBytesLeft = 0 then
        NewDisk (PExtraData(ExtraData)^, '');
      if S > Cardinal(DiskBytesLeft) then
        S := DiskBytesLeft;
      DestFile.WriteBuffer (P^, S);
      Inc (Cardinal(P), S);
      Dec (BufSize, S);
      Dec (DiskBytesLeft, S);
    end;
  end;
end;

procedure TSetupCompiler.EnumFiles (const Line: PChar; const Ext: Integer);

  function GetVersionInfo (Filename: String; var VersionInfo: TVSFixedFileInfo): Boolean;
  var
    VersionSize: Integer;
    VersionHandle: DWORD;
    VersionBuf: PChar;
    VerInfo: PVSFixedFileInfo;
    VerInfoSize: UINT;
  begin
    Result := False;

    VersionSize := GetFileVersionInfoSize(StringAsPChar(Filename),
      VersionHandle);
    if VersionSize <> 0 then begin
      GetMem (VersionBuf, VersionSize);
      try
        if GetFileVersionInfo(StringAsPChar(Filename), VersionHandle, VersionSize, VersionBuf) then begin
          if VerQueryValue(VersionBuf, '\', Pointer(VerInfo), VerInfoSize) then begin
            VersionInfo := VerInfo^;
            Result := True;
          end;
        end;
      finally
        FreeMem (VersionBuf, VersionSize);
      end;
    end;
  end;

  function EscapeBraces (const S: String): String;
  { Changes all '{' to '{{' }
  var
    I: Integer;
  begin
    Result := S;
    I := 1;
    while I <= Length(Result) do begin
      if Result[I] = '{' then begin
        Insert ('{', Result, I);
        Inc (I);
      end
      else if Result[I] in LeadBytes then
        Inc (I);
      Inc (I);
    end;
  end;

const
  ParamFilesSource = 'Source';
  ParamFilesDestDir = 'DestDir';
  ParamFilesDestName = 'DestName';
  ParamFilesCopyMode = 'CopyMode';
  ParamFilesAttribs = 'Attribs';
  ParamFilesFontInstall = 'FontInstall';
  ParamNames: array[0..10] of TParamInfo = (
    (Name: ParamCommonFlags; Flags: []),
    (Name: ParamFilesSource; Flags: [piNoEmpty, piNoQuotes]),
    (Name: ParamFilesDestDir; Flags: [piNoEmpty, piNoQuotes]),
    (Name: ParamFilesDestName; Flags: [piNoEmpty, piNoQuotes]),
    (Name: ParamFilesCopyMode; Flags: []),
    (Name: ParamFilesAttribs; Flags: []),
    (Name: ParamFilesFontInstall; Flags: [piNoEmpty]),
    (Name: ParamCommonComponents; Flags: []),
    (Name: ParamCommonTasks; Flags: []),
    (Name: ParamCommonMinVersion; Flags: []),
    (Name: ParamCommonOnlyBelowVersion; Flags: []));
  Flags: array[0..16] of PChar = (
    'confirmoverwrite', 'uninsneveruninstall', 'isreadme', 'regserver',
    'sharedfile', 'restartreplace', 'deleteafterinstall',
    'comparetimestampalso', 'fontisnttruetype', 'regtypelib', 'external',
    'skipifsourcedoesntexist', 'overwritereadonly', 'onlyifdestfileexists',
    'recursesubdirs', 'noregerror', 'allowunsafefiles');
  CopyModes: array[TSetupFileCopyMode] of PChar = (
    'normal', 'onlyifdoesntexist', 'alwaysoverwrite',
    'alwaysskipifsameorolder');
  AttribsFlags: array[0..2] of PChar = (
    'readonly', 'hidden', 'system');
var
  Params: TBreakStringArray;
  ParamNameFound: array[Low(ParamNames)..High(ParamNames)] of Boolean;
  NewFileEntry, OldFileEntry: PSetupFileEntry;
  NewFileLocationEntry: PSetupFileLocationEntry;
  NewRunEntry: PSetupRunEntry;
  P, I: Integer;
  VersionInfo: TVSFixedFileInfo;
  J: TSetupFileCopyMode;
  ValidCopyMode: Boolean;
  SourceWildcard, SourceFile, ADestDir, ADestName, AInstallFontName: String;
  ReadmeFile, ExternalFile, SourceIsWildcard, RecurseSubdirs,
    AllowUnsafeFiles: Boolean;

  procedure CheckForUnsafeFile (const Filename: String);
  { This generates errors on "unsafe files" copied to the System directory }
  begin
    if AllowUnsafeFiles or (ADestDir <> '{sys}\') then
      Exit;
    { COMCAT.DLL 5.0 }
    if (CompareText(Filename, 'COMCAT.DLL') = 0) and
       (foVersionInfoValid in NewFileLocationEntry^.Flags) and
       (NewFileLocationEntry^.FileVersionMS >= $50000) then
      AbortCompileOnLineFmt (SCompilerFilesUnsafeFile, ['COMCAT.DLL version 5.0']);
    { COMCTL32.DLL }
    if CompareText(Filename, 'COMCTL32.DLL') = 0 then
      AbortCompileOnLineFmt (SCompilerFilesUnsafeFile, ['COMCTL32.DLL']);
    { CTL3D32.DLL }
    if (CompareText(Filename, 'CTL3D32.DLL') = 0) and
       (NewFileEntry^.MinVersion.WinVersion <> 0) and
       FileSizeAndCRCIs(SourceFile, 27136, $28A66C20) then
      AbortCompileOnLineFmt (SCompilerFilesUnsafeFile, ['CTL3D32.DLL, Windows NT-specific version']);
  end;

  function Recurse (const SearchBaseDir, SearchSubDir, SearchWildcard: String): Boolean;
  { Searches for any files matching "SearchBaseDir + SearchSubDir + SearchWildcard"
    and adds FileEntries for them.
    Returns True if any files were found. }
  var
    SearchFullPath, N: String;
    SearchRec: TSearchRec;
    I: Integer;
  begin
    SearchFullPath := SearchBaseDir + SearchSubDir + SearchWildcard;
    Result := not SourceIsWildcard or
      (FindFirst(SearchFullPath, 0, SearchRec) = 0);
    if Result then begin
      try
        repeat
          if NewFileEntry = nil then begin
            NewFileEntry := AllocMem(SizeOf(TSetupFileEntry));
            SEDuplicateRec (OldFileEntry, NewFileEntry,
              SizeOf(TSetupFileEntry), SetupFileEntryStrings);
          end;
          N := '';
          if SourceIsWildcard then begin
            SourceFile := SearchBaseDir + SearchSubDir + SearchRec.Name;
            N := SearchSubDir + EscapeBraces(SearchRec.Name);
            NewFileEntry^.DestName := ADestDir + N;
          end
          else begin
            SourceFile := SearchFullPath;
            if Ext = 0 then begin
              if ADestName = '' then begin
                if not ExternalFile then begin
                  N := EscapeBraces(SearchWildcard);
                  NewFileEntry^.DestName := ADestDir + N;
                end
                else
                  { Don't append a filename to DestName on 'external' files;
                    it will be determined during installation }
                  NewFileEntry^.DestName := ADestDir;
              end
              else begin
                N := ADestName;
                NewFileEntry^.DestName := ADestDir + N;
                  { ^ user is already required to escape '{' in DestName }
                Include (NewFileEntry^.Options, foCustomDestName);
              end;
            end
            else
              NewFileEntry^.DestName := '';
          end;

          NewFileLocationEntry := nil;
          if not ExternalFile then begin
            if not DontMergeDuplicateFiles then
              for I := 0 to FileLocationEntryFilenames.Count-1 do
                { See if the source filename is already in the list of files to
                  be compressed. If so, don't duplicate it. }
                if AnsiCompareFileName(FileLocationEntryFilenames[I],
                   SourceFile) = 0 then begin
                  NewFileLocationEntry := FileLocationEntries[I];
                  NewFileEntry^.LocationEntry := I;
                  Break;
                end;
            if NewFileLocationEntry = nil then begin
              NewFileLocationEntry := AllocMem(SizeOf(TSetupFileLocationEntry));
              if CompressWithBzip then begin
                Include (NewFileLocationEntry^.Flags, foBzipped);
                Include (SetupHeader.Options, shBzipUsed);
              end;
              FileLocationEntries.Add (NewFileLocationEntry);
              FileLocationEntryFilenames.Add (SourceFile);
              NewFileEntry^.LocationEntry := FileLocationEntries.Count-1;
            end;
          end
          else begin
            NewFileEntry^.SourceFilename := SearchFullPath;
            NewFileEntry^.LocationEntry := -1;
          end;

          { Read version info }
          if not ExternalFile and (NewFileEntry^.CopyMode <> cmAlwaysOverwrite) and
             (NewFileLocationEntry^.Flags * [foVersionInfoValid, foVersionInfoNotValid] = []) then begin
            AddStatus (Format(SCompilerStatusFilesVerInfo, [SourceFile]));
            if GetVersionInfo(SourceFile, VersionInfo) then begin
              NewFileLocationEntry^.FileVersionMS := VersionInfo.dwFileVersionMS;
              NewFileLocationEntry^.FileVersionLS := VersionInfo.dwFileVersionLS;
              Include (NewFileLocationEntry^.Flags, foVersionInfoValid);
            end
            else
              Include (NewFileLocationEntry^.Flags, foVersionInfoNotValid);
          end;

          CheckForUnsafeFile (N);

          if ReadmeFile then begin
            NewRunEntry := AllocMem(Sizeof(TSetupRunEntry));
            NewRunEntry.Name := NewFileEntry.DestName;
            NewRunEntry.Components := NewFileEntry.Components;
            NewRunEntry.Tasks := NewFileEntry.Tasks;
            NewRunEntry.MinVersion := NewFileEntry.MinVersion;
            NewRunEntry.OnlyBelowVersion := NewFileEntry.OnlyBelowVersion;
            NewRunEntry.Options := [roShellExec, roSkipIfDoesntExist, roPostInstall, roSkipIfSilent];
            NewRunEntry.ShowCmd := SW_SHOWNORMAL;
            RunEntries.Insert (0, NewRunEntry);
          end;

          FileEntries.Add (NewFileEntry);
          OldFileEntry := NewFileEntry;
          NewFileEntry := nil;
        until not SourceIsWildcard or (FindNext(SearchRec) <> 0);
      finally
        if SourceIsWildcard then
          FindClose (SearchRec);
      end;
    end;
    if SourceIsWildcard and RecurseSubdirs then begin
      if FindFirst(SearchBaseDir + SearchSubDir + '*', faDirectory,
         SearchRec) = 0 then begin
        try
          repeat
            if (SearchRec.Attr and faDirectory <> 0) and
               (SearchRec.Name <> '.') and (SearchRec.Name <> '..') then
              Result := Recurse(SearchBaseDir, SearchSubDir + SearchRec.Name + '\',
                SearchWildcard) or Result;
          until FindNext(SearchRec) <> 0;
        finally
          FindClose (SearchRec);
        end;
      end;
    end;
  end;

begin
  CallIdleProc;

  FillChar (ParamNameFound, SizeOf(ParamNameFound), 0);
  if Ext = 0 then
    BreakString (Line, Params);

  NewFileEntry := nil;
  try
    NewFileEntry := AllocMem(SizeOf(TSetupFileEntry));
    with NewFileEntry^ do begin
      MinVersion := SetupHeader^.MinVersion;
      CopyMode := cmNormal;

      ADestName := '';
      ADestDir := '';
      AInstallFontName := '';
      ReadmeFile := False;
      ExternalFile := False;
      RecurseSubdirs := False;
      AllowUnsafeFiles := False;

      case Ext of
        0: for P := Low(Params) to High(Params) do
             with Params[P] do begin
               if ParamName = '' then Break;
               case CompareParamName(Params[P], ParamNames, ParamNameFound) of
                 0: while True do
                      case ExtractFlag(ParamData, Flags) of
                        -2: Break;
                        -1: AbortCompileParamError (SCompilerParamUnknownFlag2, ParamCommonFlags);
                        0: Include (Options, foConfirmOverwrite);
                        1: Include (Options, foUninsNeverUninstall);
                        2: ReadmeFile := True;
                        3: Include (Options, foRegisterServer);
                        4: Include (Options, foSharedFile);
                        5: Include (Options, foRestartReplace);
                        6: Include (Options, foDeleteAfterInstall);
                        7: Include (Options, foCompareTimeStampAlso);
                        8: Include (Options, foFontIsntTrueType);
                        9: Include (Options, foRegisterTypeLib);
                        10: ExternalFile := True;
                        11: Include (Options, foSkipIfSourceDoesntExist);
                        12: Include (Options, foOverwriteReadOnly);
                        13: Include (Options, foOnlyIfDestFileExists);
                        14: RecurseSubdirs := True;
                        15: Include (Options, foNoRegError);
                        16: AllowUnsafeFiles := True;
                      end;
                 1: SourceWildcard := ParamData;
                 2: ADestDir := ParamData;
                 3: begin
                      if AnsiPos('\', ParamData) <> 0 then
                        AbortCompileParamError (SCompilerParamNoBackslash, ParamFilesDestName);
                      ADestName := ParamData;
                    end;
                 4: begin
                      ParamData := Trim(ParamData);
                      ValidCopyMode := False;
                      for J := Low(J) to High(J) do
                        if StrIComp(CopyModes[J], StringAsPChar(ParamData)) = 0 then begin
                          CopyMode := J;
                          ValidCopyMode := True;
                          Break;
                        end;
                      if not ValidCopyMode then
                        AbortCompileParamError (SCompilerParamInvalid2, ParamFilesCopyMode);
                    end;
                 5: while True do
                      case ExtractFlag(ParamData, AttribsFlags) of
                        -2: Break;
                        -1: AbortCompileParamError (SCompilerParamUnknownFlag2, ParamFilesAttribs);
                        0: Attribs := Attribs or faReadOnly;
                        1: Attribs := Attribs or faHidden;
                        2: Attribs := Attribs or faSysFile;
                      end;
                 6: AInstallFontName := ParamData;
                 7: while True do begin
                      I := ExtractComponent(ParamData, ComponentEntries);
                      case I of
                        -2: Break;
                        -1: AbortCompileParamError (SCompilerParamUnknownComponent, ParamCommonComponents);
                        else
                         AddToCommaText(Components, PSetupComponentEntry(ComponentEntries[I]).Name)
                      end;
                    end;
                 8: while True do begin
                      I := ExtractTask(ParamData, TaskEntries);
                      case I of
                        -2: Break;
                        -1: AbortCompileParamError (SCompilerParamUnknownTask, ParamCommonTasks);
                        else
                          AddToCommaText(Tasks, PSetupTaskEntry(TaskEntries[I]).Name)
                      end;
                    end;
                 9: if not StrToVersionNumbers(ParamData, MinVersion) then
                      AbortCompileParamError (SCompilerParamInvalid2, ParamCommonMinVersion);
                 10:if not StrToVersionNumbers(ParamData, OnlyBelowVersion) then
                      AbortCompileParamError (SCompilerParamInvalid2, ParamCommonOnlyBelowVersion);
               end;
             end;
        1: begin
             SourceWildcard := 'compiler:UNINST.E32';
             FileType := ftUninstExe;
             CopyMode := cmAlwaysSkipIfSameOrOlder;
             Options := [foOverwriteSameVersion];
           end;
        2: begin
             SourceWildcard := 'compiler:REGSVR.E32';
             FileType := ftRegSvrExe;
             Options := [foUninsNeverUninstall];
           end;
      end;

      if Ext = 0 then begin
        if not ParamNameFound[1] then
          AbortCompileParamError (SCompilerParamNotSpecified, ParamFilesSource);
        if not ParamNameFound[2] then
          AbortCompileParamError (SCompilerParamNotSpecified, ParamFilesDestDir);
      end;

      if (ADestDir = '{tmp}') or (Copy(ADestDir, 1, 4) = '{tmp}\') then
        Include (Options, foDeleteAfterInstall);
      if foDeleteAfterInstall in Options then begin
        if foRestartReplace in Options then
          AbortCompileOnLineFmt (SCompilerFilesTmpBadFlag, ['restartreplace']);
        if foUninsNeverUninstall in Options then
          AbortCompileOnLineFmt (SCompilerFilesTmpBadFlag, ['uninsneveruninstall']);
        if foRegisterServer in Options then
          AbortCompileOnLineFmt (SCompilerFilesTmpBadFlag, ['regserver']);
        if foRegisterTypeLib in Options then
          AbortCompileOnLineFmt (SCompilerFilesTmpBadFlag, ['regtypelib']);
        if foSharedFile in Options then
          AbortCompileOnLineFmt (SCompilerFilesTmpBadFlag, ['sharedfile']);
        Include (Options, foUninsNeverUninstall);
      end;

      if (foRestartReplace in Options) and (foRegisterServer in Options) and
         MinVersionLessThan4(MinVersion) then
        AbortCompileOnLineFmt (SCompilerParamErrorComboUsed +
          SNewLine2 + MinVersionErrorMessage(True, $400, $400),
          [ParamCommonFlags, 'restartreplace', 'registerserver']);
      if (shAlwaysRestart in SetupHeader^.Options) or (foRestartReplace in Options) then
        HasRestart := True;
      if (foRegisterServer in Options) or (foRegisterTypeLib in Options) then
        HasRegSvr := True;

      if AInstallFontName <> '' then begin
        if not(foFontIsntTrueType in Options) then
          AInstallFontName := AInstallFontName + ' (TrueType)';
        InstallFontName := AInstallFontName;
      end;

      SourceIsWildcard := (Ext = 0) and ((Pos('*', SourceWildcard) <> 0)
        or (Pos('?', SourceWildcard) <> 0) or RecurseSubdirs);
      if ExternalFile then begin
        if RecurseSubdirs then
          AbortCompileOnLineFmt (SCompilerParamErrorBadCombo2,
            [ParamCommonFlags, 'external', 'recursesubdirs']);
        CheckDirConst (SourceWildcard, MinVersion, []);
      end;
      if (ADestName <> '') and SourceIsWildcard then
        AbortCompileOnLine (SCompilerFilesDestNameCantBeSpecified);
      CheckDirConst (ADestDir, MinVersion, []);
      ADestDir := AddBackslash(ADestDir);
      CheckDirConst (ADestName, MinVersion, []);
      if not ExternalFile then begin
        SourceWildcard := PrependSourceDirName(SourceWildcard);
        if (Ext = 0) and not SourceIsWildcard and not NewFileExists(SourceWildcard) then
          AbortCompileOnLineFmt (SCompilerSourceFileDoesntExist, [SourceWildcard]);
      end;
    end;

    if ExternalFile then
      SourceIsWildcard := False;  { don't do wildcard expansion at compile time }
    if not Recurse(ExtractFilePath(SourceWildcard), '',
       ExtractFileName(SourceWildcard)) then
      AbortCompileOnLineFmt (SCompilerFilesWildcardNotMatched, [SourceWildcard]);
  except
    SEFreeRec (NewFileEntry, SetupFileEntryStrings);
    raise;
  end;
end;

procedure UpdateTimeStamp (H: THandle);
var
  FT: TFileTime;
begin
  GetSystemTimeAsFileTime (FT);
  SetFileTime (H, nil, nil, @FT);
end;

procedure TSetupCompiler.EnumRun (const Line: PChar; const Ext: Integer);
const
  ParamRunFilename = 'Filename';
  ParamRunParameters = 'Parameters';
  ParamRunWorkingDir = 'WorkingDir';
  ParamRunRunOnceId = 'RunOnceId';
  ParamRunStatusMsg = 'StatusMsg';
  ParamDescription = 'Description';
  ParamNames: array[0..10] of TParamInfo = (
    (Name: ParamCommonFlags; Flags: []),
    (Name: ParamRunFilename; Flags: [piNoEmpty, piNoQuotes]),
    (Name: ParamRunParameters; Flags: []),
    (Name: ParamRunWorkingDir; Flags: []),
    (Name: ParamRunRunOnceId; Flags: []),
    (Name: ParamDescription; Flags: []),
    (Name: ParamCommonComponents; Flags: []),
    (Name: ParamCommonTasks; Flags: []),
    (Name: ParamCommonMinVersion; Flags: []),
    (Name: ParamCommonOnlyBelowVersion; Flags: []),
    (Name: ParamRunStatusMsg; Flags: []));
  Flags: array[0..11] of PChar = (
    'nowait', 'waituntilidle', 'shellexec', 'skipifdoesntexist',
    'runminimized', 'runmaximized', 'showcheckbox', 'postinstall',
    'unchecked', 'skipifsilent', 'skipifnotsilent', 'hidewizard');
var
  Params: TBreakStringArray;
  ParamNameFound: array[Low(ParamNames)..High(ParamNames)] of Boolean;
  NewRunEntry: PSetupRunEntry;
  P, I: Integer;
begin
  FillChar (ParamNameFound, SizeOf(ParamNameFound), 0);
  BreakString (Line, Params);

  NewRunEntry := AllocMem(SizeOf(TSetupRunEntry));
  try
    with NewRunEntry^ do begin
      MinVersion := SetupHeader^.MinVersion;
      ShowCmd := SW_SHOWNORMAL;

      for P := Low(Params) to High(Params) do
        with Params[P] do begin
          if ParamName = '' then Break;
          case CompareParamName(Params[P], ParamNames, ParamNameFound) of
            0: while True do
                 case ExtractFlag(ParamData, Flags) of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownFlag2, ParamCommonFlags);
                   0: Wait := rwNoWait;
                   1: Wait := rwWaitUntilIdle;
                   2: Include (Options, roShellExec);
                   3: Include (Options, roSkipIfDoesntExist);
                   4: ShowCmd := SW_SHOWMINNOACTIVE;
                   5: ShowCmd := SW_SHOWMAXIMIZED;
                   6: begin
                        if (Ext = 1) then
                          AbortCompileParamError (SCompilerParamUnsupportedFlag, ParamCommonFlags);
                        WarningsList.Add (Format(SCompilerRunFlagObsolete, ['showcheckbox', 'postinstall']));
                        Include (Options, roPostInstall);
                      end;
                   7: begin
                        if (Ext = 1) then
                          AbortCompileParamError (SCompilerParamUnsupportedFlag, ParamCommonFlags);
                        Include (Options, roPostInstall);
                      end;
                   8: begin
                        if (Ext = 1) then
                          AbortCompileParamError (SCompilerParamUnsupportedFlag, ParamCommonFlags);
                        Include (Options, roUnchecked);
                      end;
                   9: begin
                        if (Ext = 1) then
                          AbortCompileParamError (SCompilerParamUnsupportedFlag, ParamCommonFlags);
                        Include (Options, roSkipIfSilent);
                      end;
                   10: begin
                        if (Ext = 1) then
                          AbortCompileParamError (SCompilerParamUnsupportedFlag, ParamCommonFlags);
                        Include (Options, roSkipIfNotSilent);
                      end;
                   11: Include (Options, roHideWizard);
                 end;
            1: Name := ParamData;
            2: Parameters := ParamData;
            3: WorkingDir := ParamData;
            4: begin
                 if (Ext = 0) and (ParamData <> '') then
                   AbortCompileOnLine (SCompilerRunCantUseRunOnceId);
                 RunOnceId := ParamData;
               end;
            5: begin
                 if (Ext = 1) and (ParamData <> '') then
                   AbortCompileOnLine (SCompilerUninstallRunCantUseDescription);
                 Description := ParamData;
               end;
            6: while True do begin
                 I := ExtractComponent(ParamData, ComponentEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownComponent, ParamCommonComponents);
                   else
                     AddToCommaText(Components, PSetupComponentEntry(ComponentEntries[I]).Name)
                 end;
               end;
            7: while True do begin
                 I := ExtractTask(ParamData, TaskEntries);
                 case I of
                   -2: Break;
                   -1: AbortCompileParamError (SCompilerParamUnknownTask, ParamCommonTasks);
                   else
                     AddToCommaText(Tasks, PSetupTaskEntry(TaskEntries[I]).Name)
                 end;
               end;
            8: if not StrToVersionNumbers(ParamData, MinVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonMinVersion);
            9: if not StrToVersionNumbers(ParamData, OnlyBelowVersion) then
                 AbortCompileParamError (SCompilerParamInvalid2, ParamCommonOnlyBelowVersion);
            10: StatusMsg := ParamData;
          end;
        end;

      if not ParamNameFound[1] then
        AbortCompileParamError (SCompilerParamNotSpecified, ParamRunFilename);

      CheckDirConst (Name, MinVersion, []);
      CheckDirConst (Parameters, MinVersion, []);
      CheckDirConst (WorkingDir, MinVersion, []);
      CheckDirConst (RunOnceId, MinVersion, []);
    end;
  except
    SEFreeRec (NewRunEntry, SetupRunEntryStrings);
    raise;
  end;
  if Ext = 0 then
    RunEntries.Add(NewRunEntry)
  else
    UninstallRunEntries.Add (NewRunEntry);
end;

procedure TSetupCompiler.EnumMessages (const Line: PChar; const Ext: Integer);
var
  P, M: PChar;
  I {, L}: Integer;
  N: String;
begin
  P := StrScan(Line, '=');
  if P = nil then
    AbortCompileOnLine (SCompilerMessagesMissingEquals);
  SetLength (N, P - Line);
  Move (Line^, N[1], Length(N));
  I := GetEnumValue(TypeInfo(TSetupMessageID), 'msg' + N);
  if I = -1 then
    AbortCompileOnLineFmt (SCompilerMessagesNotRecognized, [N]);
  M := StrNew(P + 1);
  { Replace %n with actual CR/LF characters }
  P := M;
  while True do begin
    P := StrPos(P, '%n');
    if P = nil then Break;
    P[0] := #13;
    P[1] := #10;
    Inc (P, 2);
  end;
  { Replace %% with % }
  {}(* this isn't used anymore
  P := M;
  while True do begin
    P := StrPos(P, '%%');
    if P = nil then Break;
    Inc (P);
    L := StrLen(P);
    Move (P^, (P-1)^, L);
    P[L-1] := #0;
  end;*)
  if Assigned(Messages[TSetupMessageID(I)]) then begin
    StrDispose (Messages[TSetupMessageID(I)]);
    Messages[TSetupMessageID(I)] := nil;
  end;
  Messages[TSetupMessageID(I)] := M;
end;

procedure TSetupCompiler.InitLangOptions (var ALangOptions: TSetupLangOptions);
{ Initializes a TSetupLangOptions record to the default settings }
begin
  ALangOptions.LanguageName := 'English';
  ALangOptions.LanguageID := $0409;  { U.S. English }
  ALangOptions.DialogFontName := 'MS Shell Dlg';
  ALangOptions.DialogFontSize := 8;
  ALangOptions.DialogFontStandardHeight := 13;
  ALangOptions.TitleFontName := 'Arial';
  ALangOptions.TitleFontSize := 29;
  ALangOptions.WelcomeFontName := 'Verdana';
  ALangOptions.WelcomeFontSize := 12;
  ALangOptions.CopyrightFontName := 'Arial';
  ALangOptions.CopyrightFontSize := 8;
end;

procedure TSetupCompiler.ReadMessages;
var
  S, Filename, SaveAbortFilename: String;
  P: Integer;
  F: TextFile;
  I: TSetupMessageID;
begin
  { Read in all the message files }
  S := MessagesFile;
  while True do begin
    P := Pos(',', S);
    if P = 0 then P := Maxint;
    Filename := Trim(Copy(S, 1, P-1));
    if Filename = '' then
      Break;
    Filename := PrependSourceDirName(Filename);
    AddStatus (Format(SCompilerStatusReadingInFile, [Filename]));
    SaveAbortFilename := AbortFilename;
    AssignFile (F, Filename);
    FileMode := fmOpenRead or fmShareDenyWrite;  Reset (F);
    try
      AbortFilename := Filename;
      LangOptionsDirectivesSpecified := [];
      EnumIniSection (EnumLangOptions, 'LangOptions', 0, False, F);
      CallIdleProc;
      EnumIniSection (EnumMessages, 'Messages', 1, False, F);
      CallIdleProc;
    finally
      AbortFilename := SaveAbortFilename;
      CloseFile (F);
    end;
    Delete (S, 1, P);
  end;
  { Then read the [LangOptions] & [Messages] sections in the script }
  AddStatus (SCompilerStatusReadingInScriptMsgs);
  LangOptionsDirectivesSpecified := [];
  EnumIniSection (EnumLangOptions, 'LangOptions', 0, False, TextFile(nil^));
  CallIdleProc;
  EnumIniSection (EnumMessages, 'Messages', 0, False, TextFile(nil^));
  CallIdleProc;
  { Check messages }
  for I := Low(Messages) to High(Messages) do
    if Messages[I] = nil then
      AbortCompileFmt (SCompilerMessagesMissingMessage,
        [Copy(GetEnumName(TypeInfo(TSetupMessageID), Ord(I)), 4, Maxint)]);
        { ^ Copy(..., 4, Maxint) is to skip past "msg" }
  CallIdleProc;
end;

procedure TSetupCompiler.Compile;
  procedure EmptyOutputDir (const Log: Boolean);
    procedure DelFile (const Filename: String);
    begin
      if DeleteFile(OutputDir + Filename) and Log then
        AddStatus (Format(SCompilerStatusDeletingPrevious, [Filename]));
    end;
  var
    SR: TSearchRec;
    N: String;
    I: Integer;
    AllNumbers: Boolean;
  begin
    { Delete SETUP.* and SETUP-*.BIN if they existed in the output directory }
    DelFile (OutputBaseFilename + '.exe');
    DelFile ('setup.msg');
    DelFile ('setup.0');
    if FindFirst(OutputDir + OutputBaseFilename + '-*.bin', 0, SR) = 0 then begin
      repeat
        if SR.Attr and faDirectory = 0 then begin
          N := ChangeFileExt(SR.Name, '');
          AllNumbers := False;
          for I := Length(N) downto 1 do
            if N[I] in ['0'..'9'] then
              AllNumbers := True
            else begin
              if N[I] <> '-' then
                AllNumbers := False;
              Break;
            end;
          if AllNumbers then
            DelFile (SR.Name);
        end;
      until FindNext(SR) <> 0;
      FindClose (SR);
    end;
  end;
  procedure ReadTextFile (const Filename: String; var Text: String);
  var
    F: File;
    TextSize: Cardinal;
  begin
    AssignFile (F, Filename);
    FileMode := fmOpenRead or fmShareDenyWrite;  Reset (F, 1);
    try
      TextSize := FileSize(F);
      //if TextSize > 32000 then
      //  AbortCompile (SCompilerFileTooBig);
      SetLength (Text, TextSize);
      BlockRead (F, Text[1], TextSize);
    finally
      CloseFile (F);
    end;
  end;
  procedure FreeListItems (const List: TList; const NumStrings: Integer);
  var
    I: Integer;
  begin
    for I := List.Count-1 downto 0 do begin
      SEFreeRec (List[I], NumStrings);
      List.Delete (I);
    end;
  end;
  procedure FreeMessages;
  var
    I: TSetupMessageID;
  begin
    for I := High(Messages) downto Low(Messages) do begin
      StrDispose (Messages[I]);
      Messages[I] := nil;
    end;
  end;
type
  PCopyBuffer = ^TCopyBuffer;
  TCopyBuffer = array[0..32767] of Char;
var
  SetupFile: File;
  ConvertFile, ExeFile: File;
  LicenseText, InfoBeforeText, InfoAfterText: String;
  WizardImage: TMemoryStream;
  WizardSmallImage: TMemoryStream;
  BunzipDLL: TMemoryStream;

  SetupPrg: String;

  SetupLdrExeHeader: TSetupLdrExeHeader;
  SetupLdrOffsetTable: TSetupLdrOffsetTable;
  SizeOfExe, SizeOfHeaders, SizeOfMsg: Longint;

  function WriteSetup0 (var F: File): Longint;
    procedure WriteStream (Stream: TStream; var Data: TDeflateBlockWriteData);
    type
      PBuffer = ^TBuffer;
      TBuffer = array[0..8191] of Byte;
    var
      Buf: PBuffer;
      BytesLeft, Bytes: Longint;
    begin
      New (Buf);
      try
        Stream.Seek (0, soFromBeginning);
        BytesLeft := Stream.Size;
        DeflateBlockWrite (Data, BytesLeft, SizeOf(BytesLeft));
        while BytesLeft > 0 do begin
          Bytes := BytesLeft;
          if Bytes > SizeOf(Buf^) then Bytes := SizeOf(Buf^);
          Stream.ReadBuffer (Buf^, Bytes);
          DeflateBlockWrite (Data, Buf^, Bytes);
          Dec (BytesLeft, Bytes);
        end;
      finally
        Dispose (Buf);
      end;
    end;
  var
    Pos: Longint;
    J: Integer;
    Data: TDeflateBlockWriteData;
  begin
    Pos := FilePos(F);

    BlockWrite (F, SetupID, SizeOf(SetupID));

    SetupHeader^.LeadBytes := LeadBytes;
    SetupHeader^.NumTypeEntries := TypeEntries.Count;
    SetupHeader^.NumComponentEntries := ComponentEntries.Count;
    SetupHeader^.NumTaskEntries := TaskEntries.Count;
    SetupHeader^.NumDirEntries := DirEntries.Count;
    SetupHeader^.NumFileEntries := FileEntries.Count;
    SetupHeader^.NumFileLocationEntries := FileLocationEntries.Count;
    SetupHeader^.NumIconEntries := IconEntries.Count;
    SetupHeader^.NumIniEntries := IniEntries.Count;
    SetupHeader^.NumRegistryEntries := RegistryEntries.Count;
    SetupHeader^.NumInstallDeleteEntries := InstallDeleteEntries.Count;
    SetupHeader^.NumUninstallDeleteEntries := UninstallDeleteEntries.Count;
    SetupHeader^.NumRunEntries := RunEntries.Count;
    SetupHeader^.NumUninstallRunEntries := UninstallRunEntries.Count;
    SetupHeader^.LicenseText := LicenseText;
    SetupHeader^.InfoBeforeText := InfoBeforeText;
    SetupHeader^.InfoAfterText := InfoAfterText;

    DeflateBlockWriteBegin (F, InternalCompressLevel, Data);
    try
      SEDeflateBlockWrite (Data, SetupHeader^, SizeOf(TSetupHeader),
        SetupHeaderStrings);
      SEDeflateBlockWrite (Data, LangOptions, SizeOf(LangOptions),
        SetupLangOptionsStrings);

      WriteStream (WizardImage, Data);
      WriteStream (WizardSmallImage, Data);
      if shBzipUsed in SetupHeader.Options then
        WriteStream (BunzipDLL, Data);

      for J := 0 to TypeEntries.Count-1 do
        SEDeflateBlockWrite (Data, TypeEntries[J]^, SizeOf(TSetupTypeEntry),
          SetupTypeEntryStrings);
      for J := 0 to ComponentEntries.Count-1 do
        SEDeflateBlockWrite (Data, ComponentEntries[J]^, SizeOf(TSetupComponentEntry),
          SetupComponentEntryStrings);
      for J := 0 to TaskEntries.Count-1 do
        SEDeflateBlockWrite (Data, TaskEntries[J]^, SizeOf(TSetupTaskEntry),
          SetupTaskEntryStrings);
      for J := 0 to DirEntries.Count-1 do
        SEDeflateBlockWrite (Data, DirEntries[J]^, SizeOf(TSetupDirEntry),
          SetupDirEntryStrings);
      for J := 0 to FileEntries.Count-1 do
        SEDeflateBlockWrite (Data, FileEntries[J]^, SizeOf(TSetupFileEntry),
          SetupFileEntryStrings);
      for J := 0 to IconEntries.Count-1 do
        SEDeflateBlockWrite (Data, IconEntries[J]^, SizeOf(TSetupIconEntry),
          SetupIconEntryStrings);
      for J := 0 to IniEntries.Count-1 do
        SEDeflateBlockWrite (Data, IniEntries[J]^, SizeOf(TSetupIniEntry),
          SetupIniEntryStrings);
      for J := 0 to RegistryEntries.Count-1 do
        SEDeflateBlockWrite (Data, RegistryEntries[J]^, SizeOf(TSetupRegistryEntry),
          SetupRegistryEntryStrings);
      for J := 0 to InstallDeleteEntries.Count-1 do
        SEDeflateBlockWrite (Data, InstallDeleteEntries[J]^, SizeOf(TSetupDeleteEntry),
          SetupDeleteEntryStrings);
      for J := 0 to UninstallDeleteEntries.Count-1 do
        SEDeflateBlockWrite (Data, UninstallDeleteEntries[J]^, SizeOf(TSetupDeleteEntry),
          SetupDeleteEntryStrings);
      for J := 0 to RunEntries.Count-1 do
        SEDeflateBlockWrite (Data, RunEntries[J]^, SizeOf(TSetupRunEntry),
          SetupRunEntryStrings);
      for J := 0 to UninstallRunEntries.Count-1 do
        SEDeflateBlockWrite (Data, UninstallRunEntries[J]^, SizeOf(TSetupRunEntry),
          SetupRunEntryStrings);

      DeflateBlockWriteEnd (Data);
    except
      DeflateBlockWriteCancel (Data);
      raise;
    end;

    if not DiskSpanning then
      DeflateBlockWriteBegin (F, InternalCompressLevel, Data)
    else
      DeflateBlockWriteBegin (F, Z_NO_COMPRESSION, Data);
      { ^ When disk spanning is enabled, the Setup Compiler requires that
        FileLocationEntries be a fixed size, so Z_NO_COMPRESSION is used so
        that FileLocationEntries are not compressed }
    try
      for J := 0 to FileLocationEntries.Count-1 do
        DeflateBlockWrite (Data, FileLocationEntries[J]^, SizeOf(TSetupFileLocationEntry));
      DeflateBlockWriteEnd (Data);
    except
      DeflateBlockWriteCancel (Data);
      raise;
    end;

    Result := FilePos(F) - Pos;
  end;
  function CreateSetup0File: Longint;
  var
    F: File;
  begin
    AssignFile (F, OutputDir + 'setup.0');
    FileMode := fmOpenWrite or fmShareExclusive;  Rewrite (F, 1);
    try
      Result := WriteSetup0(F);
    finally
      CloseFile (F);
    end;
  end;
  procedure WriteSetupMsg (var F: File);
  type
    PMessageLengths = ^TMessageLengths;
    TMessageLengths = array[TSetupMessageID] of Smallint;
  var
    Data: TDeflateBlockWriteData;
    ZeroPad: Longint;
    Lengths: PMessageLengths;
    I: TSetupMessageID;
    Len: Integer;
    Header: TMessagesHeader;
    TotalMessagesSize: Longint;
  begin
    ZeroPad := 0;
    New (Lengths);
    try
      { Prepare the header }
      Header.CRCMessages := Longint($FFFFFFFF);  { updated below }
      TotalMessagesSize := 0;
      for I := Low(Messages) to High(Messages) do begin
        Len := StrLen(Messages[I]) + 1;
        Header.CRCMessages := UpdateCRC32(Header.CRCMessages, Messages[I]^,
          Len);
        Lengths^[I] := Len;
        Inc (TotalMessagesSize, Len);
      end;
      Header.CRCMessages := Header.CRCMessages xor Longint($FFFFFFFF);
      Header.NumMessages := Ord(High(Messages)) - Ord(Low(Messages)) + 1;
      Header.Padding := (SizeOf(MessagesHdrID) + SizeOf(Header) + SizeOf(Lengths^)) and 3;
      if Header.Padding <> 0 then
        Header.Padding := 4 - Header.Padding;
      Header.TotalSize := SizeOf(MessagesHdrID) + SizeOf(Header) + SizeOf(Lengths^) + Header.Padding + TotalMessagesSize;
      Header.NotTotalSize := not Header.TotalSize;
      Header.CRCLengths := GetCRC32(Lengths^, SizeOf(Lengths^));

      { Write it all }
      DeflateBlockWriteBegin (F, InternalCompressLevel, Data);
      try
        DeflateBlockWrite (Data, MessagesHdrID, SizeOf(MessagesHdrID));
        DeflateBlockWrite (Data, Header, SizeOf(Header));
        DeflateBlockWrite (Data, Lengths^, SizeOf(Lengths^));
        if Header.Padding <> 0 then
          DeflateBlockWrite (Data, ZeroPad, Header.Padding);
        for I := Low(Messages) to High(Messages) do
          DeflateBlockWrite (Data, Messages[I]^, Lengths^[I]);
        DeflateBlockWriteEnd (Data);
      except
        DeflateBlockWriteCancel (Data);
        raise;
      end;
    finally
      Dispose (Lengths);
    end;
  end;
  function CreateSetupMsgFile: Longint;
  var
    F: File;
  begin
    AssignFile (F, OutputDir + 'setup.msg');
    FileMode := fmOpenWrite or fmShareExclusive;  Rewrite (F, 1);
    try
      WriteSetupMsg (F);
      Result := FileSize(F);
    finally
      CloseFile (F);
    end;
  end;
  function RoundToNearestClusterSize (const L: Longint): Longint;
  begin
    Result := (L div DiskClusterSize) * DiskClusterSize;
    if L mod DiskClusterSize <> 0 then
      Inc (Result, DiskClusterSize);
  end;
  procedure CompressFiles (const FirstDestFile: String;
    const BytesToReserveOnFirstDisk: Longint);
  var
    J: Integer;
    FL: PSetupFileLocationEntry;
    Data: TExtraData;
    Date2: TFileTime;
    SizeHigh: DWORD;
    Offset: Longint;
  begin
    Data.Compiler := Self;
    Data.DestFile := nil;
    Data.CurDisk := 0;
    try
      NewDisk (Data, FirstDestFile);
      if FirstDestFile <> '' then
        Offset := Data.DestFile.Position
      else
        Offset := 0;
      if DiskSpanning then begin
        Dec (Data.DiskBytesLeft, BytesToReserveOnFirstDisk);
        if Data.DiskBytesLeft < 0 then
          AbortCompile (SCompilerNotEnoughSpaceOnFirstDisk);
      end;
      for J := 0 to FileLocationEntries.Count-1 do begin
        FL := FileLocationEntries[J];
        AddStatus (Format(SCompilerStatusFilesCompressing,
          [FileLocationEntryFilenames[J]]));
        CallIdleProc;
        Data.SourceFile := TFileStream.Create(FileLocationEntryFilenames[J],
          fmOpenRead or fmShareDenyWrite);
        try
          FL.OriginalSize := GetFileSize(Data.SourceFile.Handle, @SizeHigh);
          if (FL.OriginalSize < 0) or (SizeHigh <> 0) then
            AbortCompile (SCompilerFileLargerThan2GB);
          if Data.DiskBytesLeft < SizeOf(ZLIBID)+1 then
            NewDisk (Data, '');
          FL.FirstDisk := Data.CurDisk;
          if Data.CurDisk > 1 then
            Offset := 0;
          FL.StartOffset := Data.DestFile.Position - Offset;
          Data.DestFile.WriteBuffer (ZLIBID, SizeOf(ZLIBID));
          Dec (Data.DiskBytesLeft, SizeOf(ZLIBID));
          if not(foBzipped in FL.Flags) then begin
            CustomDeflateData (ReadProc, WriteProc, Longint(@Data), CompressLevel,
              FL.CompressedSize, FL.Adler);
          end
          else begin
            InitBzipDLL;
            BZCustomDeflateData (ReadProc, WriteProc, Longint(@Data), CompressLevel,
              FL.CompressedSize, FL.Adler);
          end;
          if FL.CompressedSize < 0 then
            { in case the file was < 2 GB before compression but > 2 GB after }
            AbortCompile (SCompilerFileLargerThan2GB);
          GetFileTime (Data.SourceFile.Handle, nil, nil, @Date2);
          FileTimeToLocalFileTime (Date2, FL.Date);
          FL.LastDisk := Data.CurDisk;
        finally
          Data.SourceFile.Free;
        end;
      end;
    finally
      EndDisk (Data);
    end;
  end;
  procedure AddSetupType(Name, Description: String; Options: TSetupTypeOptions);
  var
    NewTypeEntry: PSetupTypeEntry;
  begin
    NewTypeEntry := AllocMem(SizeOf(TSetupTypeEntry));
    NewTypeEntry.Name := Name;
    NewTypeEntry.Description := Description;
    NewTypeEntry.MinVersion := SetupHeader.MinVersion;
    NewTypeEntry.OnlyBelowVersion := SetupHeader.OnlyBelowVersion;
    NewTypeEntry.Options := Options;
    TypeEntries.Add(NewTypeEntry);
  end;
  procedure MkDirs(Dir: string);
  begin
    Dir := RemoveBackslash(Dir);
    if (Length(Dir) < 3) or DirExists(Dir) or (ExtractFilePath(Dir) = Dir) then
      Exit;
    MkDirs(ExtractFilePath(Dir));
    MkDir(Dir);
  end;
var
  I: Integer;
  SetupE32: String;
begin
  AbortLineNumber := 0;

  WizardImage := nil;
  WizardSmallImage := nil;
  BunzipDLL := nil;
  HasRestart := False;
  HasRegSvr := False;

  try
    SetupHeader := AllocMem(SizeOf(TSetupHeader));
    FillChar (SetupHeader^, SizeOf(TSetupHeader), 0);

    { Initialize defaults }
    OutputDir := 'Output';
    OutputBaseFilename := 'setup';
    InternalCompressLevel := 9;
    CompressLevel := 7;
    UseSetupLdr := True;
    DiskTotalBytes := 1457664;
    DiskClusterSize := 512;
    ReserveBytes := 0;
    with SetupHeader^.MinVersion do begin
      WinVersion := $04000000;
      NTVersion := $04000000;
    end;
    WizardStyle := wsModern;
    SetupHeader^.Options := DefaultOptionsModern;
    SetupHeader^.UninstallFilesDir := '{app}';
    SetupHeader^.BackColor := clBlue;
    SetupHeader^.BackColor2 := clBlack;
    BackSolid := False;
    SetupHeader^.WizardImageBackColor := DefaultWizardImageBackColorModern;
    SetupHeader^.WizardSmallImageBackColor := clWhite;
    SetupHeader.UninstallStyle := DefaultUninstallStyleModern;
    WizardImageFile := DefaultWizardImageFileModern;
    WizardSmallImageFile := 'compiler:WIZMODERNSMALLIMAGE.BMP';
    MessagesFile := 'compiler:DEFAULT.ISL';
    InitLangOptions (LangOptions);

    { Read [Setup] section }
    EnumIniSection (EnumSetup, 'Setup', 0, True, TextFile(nil^));
    CallIdleProc;

    { Verify settings set in [Setup] section }
    SetupHeader^.BaseFilename := OutputBaseFilename;
    if SetupDirectiveLines[ssAppName] = 0 then
      AbortCompileFmt (SCompilerEntryMissing2, ['Setup', 'AppName']);
    if SetupDirectiveLines[ssAppVerName] = 0 then
      AbortCompileFmt (SCompilerEntryMissing2, ['Setup', 'AppVerName']);
    if SetupHeader^.AppId = '' then
      SetupHeader^.AppId := SetupHeader^.AppName;
    LineNumber := SetupDirectiveLines[ssDefaultDirName];
    CheckDirConst (SetupHeader^.DefaultDirName, SetupHeader^.MinVersion, []);
    if SetupHeader^.DefaultDirName = '' then begin
      if shCreateAppDir in SetupHeader^.Options then
        AbortCompileFmt (SCompilerEntryMissing2, ['Setup', 'DefaultDirName'])
      else
        SetupHeader^.DefaultDirName := '?ERROR?';
    end;
    if MinVersionLessThan4(SetupHeader^.MinVersion) and
       ContainsBadPMChars(SetupHeader^.DefaultDirName) then
      AbortCompile (SCompilerBadCharInDefaultDirName32);
    CheckDirConst (SetupHeader^.DefaultGroupName, SetupHeader^.MinVersion, []);
    if SetupHeader^.DefaultGroupName = '' then
      SetupHeader^.DefaultGroupName := '(Default)';
    LineNumber := SetupDirectiveLines[ssUninstallDisplayIcon];
    CheckDirConst (SetupHeader^.UninstallDisplayIcon, SetupHeader^.MinVersion, []);
    LineNumber := SetupDirectiveLines[ssUninstallFilesDir];
    CheckDirConst (SetupHeader^.UninstallFilesDir, SetupHeader^.MinVersion, []);
    if BackSolid then
      SetupHeader^.BackColor2 := SetupHeader^.BackColor;
    if not DiskSpanning then begin
      DiskTotalBytes := MaxDiskSize;
      DiskClusterSize := 1;
      ReserveBytes := 0;
    end;
    LineNumber := 0;

    CompilerDir := AddBackslash(ExpandFileName(CompilerDir));
    SourceDir := AddBackslash(ExpandFileName(SourceDir));
    if IsRelativePath(OutputDir) then
      OutputDir := SourceDir + OutputDir;
    OutputDir := ExpandFileName(OutputDir);
    if not DirExists(OutputDir) then
      MkDirs (OutputDir);
    OutputDir := AddBackslash(OutputDir);
    { Clear any SETUP.* files out of the output directory first. Raise
      an error if any files still remain }
    EmptyOutputDir (True);

    { Read text files }
    if LicenseFile <> '' then begin
      AddStatus (Format(SCompilerStatusReadingFile, ['LicenseFile']));
      ReadTextFile (PrependSourceDirName(LicenseFile), LicenseText);
    end;
    if InfoBeforeFile <> '' then begin
      AddStatus (Format(SCompilerStatusReadingFile, ['InfoBeforeFile']));
      ReadTextFile (PrependSourceDirName(InfoBeforeFile), InfoBeforeText);
    end;
    if InfoAfterFile <> '' then begin
      AddStatus (Format(SCompilerStatusReadingFile, ['InfoAfterFile']));
      ReadTextFile (PrependSourceDirName(InfoAfterFile), InfoAfterText);
    end;
    CallIdleProc;

    { Read wizard image }
    AddStatus (Format(SCompilerStatusReadingFile, ['WizardImage']));
    AddStatus (Format(SCompilerStatusReadingInFile, [PrependSourceDirName(WizardImageFile)]));
    WizardImage := TMemoryStream.Create;
    WizardImage.LoadFromFile (PrependSourceDirName(WizardImageFile));
    AddStatus (Format(SCompilerStatusReadingFile, ['WizardSmallImage']));
    AddStatus (Format(SCompilerStatusReadingInFile, [PrependSourceDirName(WizardSmallImageFile)]));
    WizardSmallImage := TMemoryStream.Create;
    WizardSmallImage.LoadFromFile (PrependSourceDirName(WizardSmallImageFile));

    { Read bunzip DLL }
    AddStatus (Format(SCompilerStatusReadingFile, ['isbunzip.dll']));
    BunzipDLL := TMemoryStream.Create;
    BunzipDLL.LoadFromFile (CompilerDir + 'isbunzip.dll');

    { Read messages & [LangOptions] }
    AddStatus (SCompilerStatusReadingMessages);
    ReadMessages;

    { Read [Types] section }
    EnumIniSection (EnumTypes, 'Types', 0, True, TextFile(nil^));
    CallIdleProc;

    { Read [Components] section }
    EnumIniSection (EnumComponents, 'Components', 0, True, TextFile(nil^));
    CallIdleProc;

    { Read [Tasks] section }
    EnumIniSection (EnumTasks, 'Tasks', 0, True, TextFile(nil^));
    CallIdleProc;

    { Read [Dirs] section }
    EnumIniSection (EnumDirs, 'Dirs', 0, True, TextFile(nil^));
    CallIdleProc;

    { Read [Icons] section }
    EnumIniSection (EnumIcons, 'Icons', 0, True, TextFile(nil^));
    CallIdleProc;

    { Read [INI] section }
    EnumIniSection (EnumINI, 'INI', 0, True, TextFile(nil^));
    CallIdleProc;

    { Read [Registry] section }
    EnumIniSection (EnumRegistry, 'Registry', 0, True, TextFile(nil^));
    CallIdleProc;

    { Read [InstallDelete] section }
    EnumIniSection (EnumDelete, 'InstallDelete', 0, True, TextFile(nil^));
    CallIdleProc;

    { Read [UninstallDelete] section }
    EnumIniSection (EnumDelete, 'UninstallDelete', 1, True, TextFile(nil^));
    CallIdleProc;

    { Read [Run] section }
    EnumIniSection (EnumRun, 'Run', 0, True, TextFile(nil^));
    CallIdleProc;

    { Read [UninstallRun] section }
    EnumIniSection (EnumRun, 'UninstallRun', 1, True, TextFile(nil^));
    CallIdleProc;

    { Read [Files] section }
    if shUninstallable in SetupHeader^.Options then
      EnumFiles ('', 1);
    EnumIniSection (EnumFiles, 'Files', 0, True, TextFile(nil^));
    if HasRestart and HasRegSvr then
      { include REGSVR.E32 }
      EnumFiles ('', 2);
    CallIdleProc;

    { Add default types if necessary }
    if (ComponentEntries.Count > 0) and (TypeEntries.Count = 0) then begin
      AddSetupType(DefaultTypeEntryNames[0], Messages[msgFullInstallation], []);
      AddSetupType(DefaultTypeEntryNames[1], Messages[msgCompactInstallation], []);
      AddSetupType(DefaultTypeEntryNames[2], Messages[msgCustomInstallation], [toIsCustom]);
    end;

    { Create setup files }
    AddStatus (SCompilerStatusCreateSetupFiles);

    if WizardStyle = wsClassic then
      SetupE32 := 'SETUPCLASSIC.E32'
    else
      SetupE32 := 'SETUPMODERN.E32';

    ExeFilename := OutputDir + OutputBaseFilename + '.exe';
    SetupPrg := CompilerDir;
    if not UseSetupLdr then
      SetupPrg := SetupPrg + SetupE32
    else
      SetupPrg := SetupPrg + 'SETUPLDR.E32';
    if not FileCopy(SetupPrg, ExeFilename, False, fmOpenRead or fmShareDenyWrite) then
      AbortCompileFmt (SCompilerCopyError2, [SetupPrg, ExeFilename]);
    FileSetAttr (ExeFilename, faArchive);  { if there was a read-only attribute, remove it }
    AssignFile (SetupFile, ExeFilename);
    FileMode := fmOpenRead or fmShareDenyWrite;  Reset (SetupFile, 1);
    try
      SizeOfExe := FileSize(SetupFile);
    finally
      CloseFile (SetupFile);
    end;
    CallIdleProc;
    try
      if not UseSetupLdr then begin
        { Create SETUP.MSG in output dir }
        SizeOfMsg := CreateSetupMsgFile;

        if not DiskSpanning then begin
          { Create SETUP.0 and SETUP-1.BIN }
          CompressFiles ('', 0);
          CreateSetup0File;
        end
        else begin
          { Create SETUP.0 and SETUP-*.BIN }
          SizeOfHeaders := CreateSetup0File;
          CompressFiles ('', RoundToNearestClusterSize(SizeOfExe) +
            RoundToNearestClusterSize(SizeOfHeaders) +
            RoundToNearestClusterSize(SizeOfMsg) +
            RoundToNearestClusterSize(ReserveBytes));
          { CompressFiles modifies setup header data, so go back and
            rewrite it }
          if CreateSetup0File <> SizeOfHeaders then
            { Make sure new and old size match. No reason why they
              shouldn't but check just in case }
            AbortCompile (SCompilerSetup0Mismatch);
        end;
      end
      else begin
        { When disk spanning isn't used, place the compressed files inside
          SETUP.EXE } 
        if not DiskSpanning then
          CompressFiles (ExeFilename, 0);

        AssignFile (ExeFile, ExeFilename);
        FileMode := fmOpenReadWrite or fmShareExclusive;  Reset (ExeFile, 1);
        try
          Seek (ExeFile, FileSize(ExeFile));

          { Move the data from SETUP.E?? into the SETUP.EXE, and write
            header data }
          SetupLdrOffsetTable.ID := SetupLdrOffsetTableID;
          SetupLdrOffsetTable.OffsetEXE := FilePos(ExeFile);
          AssignFile (ConvertFile, CompilerDir + SetupE32);
          FileMode := fmOpenRead or fmShareDenyWrite;  Reset (ConvertFile, 1);
          try
            SetupLdrOffsetTable.UncompressedSizeEXE := FileSize(ConvertFile);
            DeflateData (ConvertFile, SetupLdrOffsetTable.UncompressedSizeEXE, True,
              ExeFile, -1, True, InternalCompressLevel,
              SetupLdrOffsetTable.CompressedSizeEXE, SetupLdrOffsetTable.AdlerEXE);
          finally
            CloseFile (ConvertFile);
          end;
          SetupLdrOffsetTable.OffsetMsg := FilePos(ExeFile);
          WriteSetupMsg (ExeFile);
          SetupLdrOffsetTable.Offset0 := FilePos(ExeFile);
          SizeOfHeaders := WriteSetup0(ExeFile);
          SetupLdrOffsetTable.TotalSize := FileSize(ExeFile) + SizeOf(SetupLdrOffsetTable);
          if DiskSpanning then begin
            SetupLdrOffsetTable.Offset1 := 0;
            { Compress the files in SETUP-*.BIN after we know the size of
              SETUP.EXE }
            CompressFiles ('',
              RoundToNearestClusterSize(SetupLdrOffsetTable.TotalSize) +
              RoundToNearestClusterSize(ReserveBytes));
            { CompressFiles modifies setup header data, so go back and
              rewrite it }
            Seek (ExeFile, SetupLdrOffsetTable.Offset0);
            if WriteSetup0(ExeFile) <> SizeOfHeaders then
              { Make sure new and old size match. No reason why they
                shouldn't but check just in case }
              AbortCompile (SCompilerSetup0Mismatch);
          end
          else
            SetupLdrOffsetTable.Offset1 := SizeOfExe;

          { Write SetupLdrExeHeader and SetupLdrOffsetTable to SETUP.EXE }
          with SetupLdrExeHeader do begin
            ID := SetupLdrExeHeaderID;
            OffsetTableOffset := FileSize(ExeFile);
            NotOffsetTableOffset := not OffsetTableOffset;
          end;
          Seek (ExeFile, SetupLdrExeHeaderOffset);
          BlockWrite (ExeFile, SetupLdrExeHeader, SizeOf(SetupLdrExeHeader));
          Seek (ExeFile, SetupLdrExeHeader.OffsetTableOffset);
          BlockWrite (ExeFile, SetupLdrOffsetTable, SizeOf(SetupLdrOffsetTable));

          { For some reason, on Win95 the date/time of the EXE sometimes
            doesn't get updated after it's been written to so it has to
            manually set it. (I don't get it!!) }
          UpdateTimeStamp (TFileRec(ExeFile).Handle);
        finally
          CloseFile (ExeFile);
        end;
      end;
    except
      EmptyOutputDir (False);
      raise;
    end;
    CallIdleProc;

    { Done }
    AddStatus ('');
    for I := 0 to WarningsList.Count-1 do
      AddStatus (SCompilerStatusWarning + WarningsList[I]);
    asm jmp @1; db 0,'Inno Setup Compiler, Copyright (C) 1998-2001 Jordan Russell',0; @1: end;
    { Note: Removing or modifying the copyright text is a violation of the
      Inno Setup license agreement; see LICENSE.TXT. }
  finally
    WarningsList.Clear;
    { Free all the data }
    SEFreeRec (SetupHeader, SetupHeaderStrings);
    BunzipDLL.Free;
    WizardSmallImage.Free;
    WizardImage.Free;
    FreeListItems (TypeEntries, SetupTypeEntryStrings);
    FreeListItems (ComponentEntries, SetupComponentEntryStrings);
    FreeListItems (TaskEntries, SetupTaskEntryStrings);
    FreeListItems (DirEntries, SetupDirEntryStrings);
    FreeListItems (FileEntries, SetupFileEntryStrings);
    FreeListItems (FileLocationEntries, SetupFileLocationEntryStrings);
    FreeListItems (IconEntries, SetupIconEntryStrings);
    FreeListItems (IniEntries, SetupIniEntryStrings);
    FreeListItems (RegistryEntries, SetupRegistryEntryStrings);
    FreeListItems (InstallDeleteEntries, SetupDeleteEntryStrings);
    FreeListItems (UninstallDeleteEntries, SetupDeleteEntryStrings);
    FreeListItems (RunEntries, SetupRunEntryStrings);
    FreeListItems (UninstallRunEntries, SetupRunEntryStrings);
    FileLocationEntryFilenames.Clear;
    FreeMessages;
  end;
end;


{ Interface functions }

function ISCompileScript (const Params: TCompileScriptParams;
  const PropagateExceptions: Boolean): Integer;
var
  SetupCompiler: TSetupCompiler;
  Data: TCompilerCallbackData;
  S: String;
begin
  if (Params.Size <> SizeOf(Params)) or
     not Assigned(Params.CallbackProc) then begin
    Result := isceInvalidParam;
    Exit;
  end;
  SetupCompiler := TSetupCompiler.Create(nil);
  try
    SetupCompiler.AppData := Params.AppData;
    SetupCompiler.CallbackProc := Params.CallbackProc;
    if Assigned(Params.CompilerPath) then
      SetupCompiler.CompilerDir := Params.CompilerPath
    else
      SetupCompiler.CompilerDir := ExtractFilePath(GetSelfFilename);
    SetupCompiler.SourceDir := Params.SourcePath;

    Result := isceNoError;
    try
      SetupCompiler.Compile;
    except
      on E: Exception do begin
        Result := isceCompileFailure;
        Data.ErrorMsg := nil;
        Data.ErrorFilename := nil;
        Data.ErrorLine := 0;
        if not(E is EAbort) then begin
          S := AddPeriod(E.Message);
          Data.ErrorMsg := PChar(S);
        end;
        if E is EISCompileError then begin
          if EISCompileError(E).Filename <> '' then
            Data.ErrorFilename := PChar(EISCompileError(E).Filename);
          Data.ErrorLine := EISCompileError(E).LineNumber;
        end;
        Params.CallbackProc (iscbNotifyError, Data, Params.AppData);
        if PropagateExceptions then
          raise;
        Exit;
      end;
    end;
    Data.OutputExeFilename := PChar(SetupCompiler.ExeFilename);
    Params.CallbackProc (iscbNotifySuccess, Data, Params.AppData);
  finally
    SetupCompiler.Free;
  end;
end;

function ISGetVersion: PCompilerVersionInfo;
const
  Ver: TCompilerVersionInfo =
   (Title: SetupTitle; Version: SetupVersion; BinVersion: SetupBinVersion);
begin
  Result := @Ver;
end;

end.
